# ---START TRANSPORT FUNCTIONS---

function calcTransport(tracerData::JZTracGridType,metState::JZMetType,
                        pGridStart::JZPGridType,pGridEnd::JZPGridType,
                        timeData::JZTimeType,gridData::JZGridType,
                        constData::JZConstDataType)
    # FUNCTION NAME     : calcTransport
    # DESCRIPTION       : Run the TPCORE transport routines
    #                     adapated from those in GEOS-Chem
    # INPUTS:
    #   tracerData      : Tracer data
    #   metState        : Meteorological state variable
    #   pGridStart      : Pressure grid data at the start
    #                     of the dynamic timestep
    #   pGridEnd        : Pressure grid data at the end
    #                     of the dynamic timestep
    #   timeData        : Timing data
    #   gridData        : Resolution specification. Includes
    #                     advection factors
    #   constData       : Important constants (radius of Earth etc)
    # OUTPUTS:
    #   [Modified tracerData]
    #
    # ---START CODE---
    # Yet to decide if this is worth doing - recalculate basis surface pressure grids
    pSurfStart  = dfill(0.0,(gridData.nLon,gridData.nLat),workers(),gridData.chunk2D)
    pSurfEnd    = dfill(0.0,(gridData.nLon,gridData.nLat),workers(),gridData.chunk2D)
    pEdgeStart  = pGridStart.pEdge
    pEdgeEnd    = pGridEnd.pEdge
    @sync begin
        [@spawnat iWorker begin
            p2DS = localpart(pSurfStart)
            p2DE = localpart(pSurfEnd)
            p3DS = localpart(pEdgeStart)
            p3DE = localpart(pEdgeEnd)
            p2DS[:] = p3DS[:,:,1]
            p2DE[:] = p3DE[:,:,1]
        end
        for iWorker in workers()]
    end

    # Air and tracer mass immediately before advection
    airMassStart = calcAirMass(pGridStart,gridData,constData)

    tMassStart = dfill(0.0,(gridData.nLon,gridData.nLat,gridData.nLev,tracerData.nTracers),
                        workers(),[gridData.chunk3D...,1])
    mixRatioStart = tracerData.mixRatio
    relWeight = tracerData.relWeight

    @sync begin
        [@spawnat iWorker begin
            mixRatio = localpart(mixRatioStart)
            tMass = localpart(tMassStart)
            airMass = localpart(airMassStart)
            relWeight = remotecall_fetch(myid(),()->relWeight)
            for iTrc in 1:size(mixRatio,4)
                tMass[:,:,:,iTrc] = mixRatio[:,:,:,iTrc] .* airMass[:,:,:] ./ relWeight[iTrc]
            end
        end
        for iWorker in workers()]
    end
   
    ## DEBUG
    #@printf("Surface Ap/Bp: %10.4e:%10.4e\n",gridData.deltaPOff[1],gridData.deltaPFac[1])
    #@printf("Cap factor: %10.4e\n",gridData.capFactor)

    # pSurfStart = Timestep start pSurf
    # pSurfEnd = Timestep end pSurf

    # Calculate air mass fluxes (with pressure fixer)
    (lonFlux,latFlux,vertFlux,dpGridFlux) = calcMassFlux(   copy(pSurfStart),   copy(pSurfEnd),
                                                            metState,           gridData            )

    ## DEBUG
    #@printf("Fixed lonFlux/latFlux: %10.4e:%10.4e\n",lonFlux[20,20,1],latFlux[20,20,1])

    # This is optional in the main code, but the controlling switch is hard-coded
    # Overwrite grid flux
    @sync begin
        [@spawnat iWorker begin
            p2DS = localpart(pSurfStart)
            p2DE = localpart(pSurfEnd)
            cdpGF = localpart(dpGridFlux)
            cdpGF[:] = p2DE[:] - p2DS[:]
        end
        for iWorker in workers()]
    end

#    # This is unnecessary when the above lines are hardcoded, but code left in for future users
#    @sync begin
#        [@spawnat iWorker begin
#            p2DS = localpart(pSurfStart)
#            p2DE = localpart(pSurfEnd)
#            cdpGF = localpart(dpGridFlux)
#            p2DE[:] = p2DS[:] + cdpGF[:]
#        end
#        for iWorker in workers()]
#    end

    # What follows is equivalent to TPCORE_FVDAS from GEOS-Chem
    # However, in GEOS-Chem the vertical indexing for
    #   UWind
    #   VWind
    #   tracerGrid
    #   lonFlux
    #   latFlux
    # are reversed.

    # For safety's sake, re-average the polar caps
    avgPole(pSurfStart,gridData)
    avgPole(pSurfEnd,gridData)

    # At this point, tpcore_fvdas recalculates geoFactor/capFactor - I see no reason to do so
    # Use recommendations for IORD, JORD and KORD
    lonOrd = 3
    latOrd = 3
    levOrd = 7
    if lonOrd <= 0
        if (nLon >= 144)
            lonLim = 0
        elseif (nLon >= 72)
            lonLim = 1
        else
            lonLim = 2
        end
    else
        lonLim = lonOrd - 3
    end
    if latOrd <= 0
        if nLat >= 91
            latLim = 0
        elseif nLat >= 46
            latLim = 1
        else
            latLim = 2
        end
    else
        latLim = latOrd - 3
    end

    levLim = max(levOrd-3,0)

    # tpcore_fvdas recalculates dpOff and dpFac... no reason to do so
    # Next: Set_Press_Terms, called once for each level
    # (Line 620->2082 in tpcore_fvdas_mod.F90)
    (dpStart,dpMid,dpShift) = calcPTerms(pSurfStart,pSurfEnd,gridData)

    ## DEBUG
    #@printf("dpStart/Mid/Shift at [20,20,1]: %10.4e:%10.4e:%10.4e\n",dpStart[20,20,1],dpMid[20,20,1],dpShift[20,20,1])

    # --- CHECKPOINT: VALIDATED ---
    tic()
    lastTime = 0.0

    # Average tracer concentrations over the poles
    avgTracerPole(tracerData.mixRatio,dpStart,gridData)
    
    # Calculate Courant-Friedrichs-Lewy numbers based onf fluxes
    nLon = gridData.nLon
    nLat = gridData.nLat
    nLev = gridData.nLev
    nCap = gridData.nCap
    UCourant = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    VCourant = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)

    cosE = gridData.cosLatEdge
    @sync begin
        [@spawnat iWorker begin
            localUC = localpart(UCourant)
            localVC = localpart(VCourant)
            localDPM = localpart(dpMid)
            localLonF = localpart(lonFlux)
            localLatF = localpart(latFlux)
            localDPSh = localpart(dpShift)
            cosEdge = remotecall_fetch(myid(),()->cosE)
            localIdx = myindexes(UCourant)
            nCap = remotecall_fetch(myid(),()->nCap)
            nLat = remotecall_fetch(myid(),()->nLat)
            localLat = localIdx[2]
            goodLat = (nCap+1):(nLat - nCap)
            for iLat in 1:length(localLat)
                targLat = localLat[iLat]
                if targLat > nCap && targLat < (nLat + 1 - nCap)
                    # Calculate horizontal flux first
                    if targLat < (nLat - nCap)
                        localUC[:,iLat,:] = localLonF[:,iLat,:] ./ localDPSh[:,iLat,:]
                    end
                    # Do we need to communicate with another worker for the previous latitude's data?
                    if iLat == 1
                        localVC[:,iLat,:] = localLatF[:,iLat,:] ./ *(0.5.*cosEdge[iLat].*
                            (localDPM[:,iLat,:] + convert(Array,dpMid[localLon,targLat-1,:])))
                    else
                        localVC[:,iLat,:] = localLatF[:,iLat,:] ./ *(0.5.*cosEdge[iLat].*
                            sum(localDPM[:,(iLat-1):iLat,:],2))
                    end
                end
            end
        end
        for iWorker in workers()]
    end

    # GEOS-Chem reruns the divergence calculation here; no need
    # Emulating Set_Cross_Terms (averaged courant numbers)
    UCAvg = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    VCAvg = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    vertMassFlux = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    dpFac = gridData.deltaPFac

    @sync begin
        [@spawnat iWorker begin
            cUCA = localpart(UCAvg)
            cVCA = localpart(VCAvg)
            cUC = localpart(UCourant)
            cVC = localpart(VCourant)
            localIdx = myindexes(UCAvg)
            localLon = localIdx[1]
            localLat = localIdx[2]
            nLon = remotecall_fetch(myid(),()->nLon)
            nCap = remotecall_fetch(myid(),()->nCap)
            nLat = remotecall_fetch(myid(),()->nLat)
            # Horizontal first
            for iLon = 1:length(localLon)
                if iLon < length(localLon)
                    # Calculate internally
                    cUCA[iLon,:,:] = 0.5.*(cUC[iLon,:,:] + cUC[iLon+1,:,:])
                else
                    # Communicate with another worker
                    targLon = localLon[end] + 1
                    if targLon > nLon
                        targLon = 1
                    end
                    cUCA[iLon,:,:] = 0.5.*(cUC[iLon,:,:] + convert(Array,UCourant[targLon,localLat,:]))
                end
            end
            # Now vertical
            for iLat = 1:length(localLat)
                # Ignore extremes
                if localLat[iLat] > 1 && localLat[iLat] < nLat
                    if iLat < length(localLat)
                        # Local calculation only
                        cVCA[:,iLat,:] = 0.5.*(cVC[:,iLat,:] + cVC[:,iLat+1,:])
                    else
                        # Communication required
                        targLat = localLat[iLat] + 1
                        cVCA[:,iLat,:] = 0.5.*(cVC[:,iLat,:] + convert(Array,VCourant[localLon,targLat,:]))
                    end
                end
            end
            # Poles need special handling if the cap is not enlarged!
            if nCap == 1 || nCap == 1
                error("Code for unenlarged polar cap is not in place")
            end
        end
        for iWorker in workers()]

        # TPCore recalculates dpGridFlux here (as dps_ctm) - no reason to do so

        # Vertical mass flux calculated by conservation
        # This is the mass flux from the layer below [?]
        [@spawnat iWorker begin
            cvertFlux = localpart(vertFlux)
            dBk = remotecall_fetch(myid(),()->dpFac)
            cdpGrid = localpart(dpGridFlux)
            cvertMassFlux = localpart(vertMassFlux)
            cvertMassFlux[:,:,1] = 0.0
            cvertMassFlux[:,:,end] = cvertFlux[:,:,end] - (dBk[end].*cdpGrid)
            for iLev in ((length(dBk)-1):-1:2)
                cvertMassFlux[:,:,iLev] = cvertMassFlux[:,:,iLev+1] + 
                    (cvertFlux[:,:,iLev] - (dBk[iLev].*cdpGrid))
            end
        end
        for iWorker in workers()]
    end

    # Determine the latitude range outside of which we need to use the flux-form 
    # semi-Lagrangian scheme for each layer
    minTPLat = zeros(nLev)
    maxTPLat = zeros(nLev)
    @sync begin
        latRange = ones(2,nLev,nworkers())
        [latRange[:,:,iWorker] = @fetchfrom workers()[iWorker] begin
            localIdx = myindexes(UCourant)
            localLat = localIdx[2]
            nLat = remotecall_fetch(myid(),()->nLat)
            nLev = length(localIdx[3])
            miniRange = fill(1,2,nLev)
            miniRange[2,:] = nLat
            midLat = itrunc((nLat + 1)/2)
            evalNH = localLat[end] >= midLat
            evalSH = localLat[1] <= midLat
            if evalSH || evalNH
                cUC = localpart(UCourant)
                nCap = remotecall_fetch(myid(),()->nCap)
                miniRange[2,:] = nLat - nCap
                miniRange[1,:] = nCap + 1
                iLatStart = findfirst(localLat.>=midLat)
                for iLev = 1:nLev
                    # Determine the Northern limit
                    iLat = iLatStart
                    while evalNH
                        highUC = any(abs(cUC[:,iLat,iLev]) .> 1.0)
                        if highUC
                            miniRange[2,iLev] = iLat - 1
                            evalNH = false
                        else
                            evalNH = iLat < length(localLat)
                            iLat += 1
                        end
                    end
                    iLat = iLatStart
                    while evalSH
                        highUC = any(abs(cUC[:,iLat,iLev]) .> 1.0)
                        if highUC
                            miniRange[1,iLev] = iLat + 1
                            evalSH = false
                        else
                            evalSH = iLat > 1
                            iLat -= 1
                        end
                    end
                end
            end
            miniRange
        end
        for iWorker in 1:nworkers()]
    end

    latRangeSub = ones(2,nLev)
    # Highest valid SH band and lowest valid NH band
    latRangeSub[1,:] = maximum(latRange[1,:,:],3)
    latRangeSub[2,:] = minimum(latRange[2,:,:],3)

    #@printf("Surface valid latitude range: %2i -- %2i\n",latRangeSub[1,1],latRangeSub[2,1])
    #@printf("Top valid latitude range    : %2i -- %2i\n",latRangeSub[1,end],latRangeSub[2,end])

    ## DEBUG
    #@printf("UC and VC at [20,20,1]: %10.4e:%10.4e\n",UCourant[20,20,1],VCourant[20,20,1])

    # This loop is relevant to ADVEC_CONSRV_OPT = 2 in subroutine TPCORE_FVDAS
    # Essentially repeating calcPTerms but for the end point
    dpEnd = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    dpOff = gridData.deltaPOff
    dpFac = gridData.deltaPFac
    pOffset = gridData.pOffset

    # Recalculate surface pressure for... some reason?
    pSurfEndAlt = dfill(0.0,(nLon,nLat),workers(),gridData.chunk2D)
    @sync begin
        [@spawnat iWorker begin
            dAp = remotecall_fetch(myid(),()->dpOff)
            dBk = remotecall_fetch(myid(),()->dpFac)
            nLev = remotecall_fetch(myid(),()->nLev)

            # Core copy of 3D output grid
            cdpEnd = localpart(dpEnd)

            # Core copy of 2D pressure grids
            pSE = localpart(pSurfEnd)
            pSEA = localpart(pSurfEndAlt)
            for iLev = 1:nLev
                cdpEnd[:,:,iLev] = dAp[iLev] + (dBk[iLev].*pSE[:])
            end
            # Add the pressure difference in each level to get the surface pressure
            # Add the pressure offset at the top of the grid
            Ap = remotecall_fetch(myid(),()->pOffset)
            pSEA[:] = Ap[end] + sum(cdpEnd,3)
        end
        for iWorker in workers()]
    end

    # The relative changes are ~1e-13 hPa - negligible
    mixRatioNew = copy(mixRatioStart)
    nTrc = tracerData.nTracers
    qqLonFull = dfill(0.0,(nLon,nLat,nLev,nTrc),workers(),[gridData.chunk3D...,1])
    qqLatFull = dfill(0.0,(nLon,nLat,nLev,nTrc),workers(),[gridData.chunk3D...,1])
    dMRStartFull = dfill(0.0,(nLon,nLat,nLev,nTrc),workers(),[gridData.chunk3D...,1]) # dq1
    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(mixRatioStart)
            localLon = localIdx[1]
            nLonLocal = length(localLon)
            localLat = localIdx[2]
            nLatLocal = length(localLat)
            cmixRatio = localpart(mixRatioStart)
            cdpStart = localpart(dpStart)
            cUCA = localpart(UCAvg)
            cVCA = localpart(VCAvg)
            nTrc = size(cmixRatio,4)
            nLev = size(cmixRatio,3)
            nLon = remotecall_fetch(myid(),()->nLon)
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            latRange = remotecall_fetch(myid(),()->latRangeSub)
            cqqLonFull = localpart(qqLonFull)
            cqqLatFull = localpart(qqLatFull)
            # Each core has a full complement of levels and tracers
            for iTrc in 1:nTrc
                for iLev in 1:nLev
                    minValidS = latRange[1,iLev]
                    maxValidN = latRange[2,iLev]
                    cmixRatioMini = cmixRatio[:,:,iLev,iTrc]
                    dMRStart = localpart(dMRStartFull)
                    dMRStart[:,:,iLev,iTrc] = cmixRatioMini[:,:,1,1] .* cdpStart[:,:,iLev]
                    qqLon = zeros(nLonLocal,nLatLocal)
                    qqLat = zeros(nLonLocal,nLatLocal)
                    # Calculate advection cross terms
                    for iLat = 1:nLatLocal
                        targLat = localLat[iLat]
                        # Ignore polar caps completely for now
                        if targLat > nCap && targLat <= (nLat-nCap)
                            if targLat < minValidS || targLat > maxValidN
                                # CFL > 1
                                for iLon = 1:nLonLocal
                                    targLon = localLon[iLon]
                                    fCFL = cUCA[iLon,iLat,iLev] # True average CFL
                                    iCFL = itrunc(fCFL)         # Integer part of average CFL
                                    rCFL = fCFL - iCFL          # Remainder
                                    tLon = iLon - iCFL          # Target cell
                                    if tLon < 1 || tLon > nLonLocal
                                        # Data needed from another worker
                                        tLonOut = 1 + mod(targLon - (iCFL+1),nLon)
                                        qt0 = mixRatioStart[tLonOut,targLat,iLev,iTrc]
                                    else
                                        qt0 = cmixRatioMini[tLon,iLat]
                                    end
                                    if fCFL >= 0.0
                                        tLon = tLon - 1
                                        if tLon < 1 || tLon > nLonLocal
                                            # Data needed from another worker
                                            tLonOut = 1 + mod(targLon - (iCFL+2),nLon)
                                            qtM1 = mixRatioStart[tLonOut,targLat,iLev,iTrc]
                                        else
                                            qtM1 = cmixRatioMini[tLon,iLat]
                                        end
                                        qtDiff = qtM1 - qt0
                                    else
                                        tLon = tLon + 1
                                        if tLon < 1 || tLon > nLonLocal
                                            # Data needed from another worker
                                            tLonOut = 1 + mod(targLon - (iCFL),nLon)
                                            qtP1 = mixRatioStart[tLonOut,targLat,iLev,iTrc]
                                        else
                                            qtP1 = cmixRatioMini[tLon,iLat]
                                        end
                                        qtDiff = qt0 - qtP1
                                    end
                                    qqLon[iLon,iLat] = qt0 + (rCFL*qtDiff) - cmixRatioMini[iLon,iLat]
                                end
                            else
                                # CFL < 1
                                for iLon = 1:nLonLocal
                                    targLon = localLon[iLon]
                                    avgCourant = cUCA[iLon,iLat,iLev]
                                    tLon = itrunc(iLon-avgCourant) # Target cell
                                    if tLon < 1 || tLon > nLonLocal
                                        # Communication req'd
                                        tLonOut = 1 + mod(itrunc(targLon-avgCourant)-1,nLon)
                                        qt0 = mixRatioStart[tLonOut,iLat,iLev,iTrc]
                                    else
                                        qt0 = cmixRatioMini[tLon,iLat]
                                    end
                                    tLon = tLon + 1 # Target cell
                                    if tLon < 1 || tLon > nLonLocal
                                        # Communication req'd
                                        tLonOut = 1 + mod(itrunc(targLon-avgCourant)-1,nLon)
                                        qtP1 = mixRatioStart[tLonOut,targLat,iLev,iTrc]
                                    else
                                        qtP1 = cmixRatioMini[tLon,iLat]
                                    end
                                    qtDiff = qt0 - qtP1
                                    qqLon[iLon,iLat] = avgCourant * qtDiff
                                end
                            end
                            # Now vertical flux
                            for iLon = 1:nLonLocal
                                avgCourant = cVCA[iLon,iLat,iLev]
                                targLon = localLon[iLon]
                                tLat = itrunc(iLat-avgCourant)
                                if tLat < 1 || tLat > nLatLocal
                                    tLatOut = itrunc(targLat-avgCourant)
                                    qt0 = mixRatioStart[targLon,tLatOut,iLev,iTrc]
                                else
                                    qt0 = cmixRatioMini[iLon,tLat]
                                end
                                tLat += 1
                                if tLat < 1 || tLat > nLatLocal
                                    tLatOut = itrunc(targLat+1-avgCourant)
                                    qtP1 = mixRatioStart[targLon,tLatOut,iLev,iTrc]
                                else
                                    qtP1 = cmixRatioMini[iLon,tLat]
                                end
                                qtDiff = qt0 - qtP1
                                qqLat[iLon,iLat] = avgCourant * qtDiff
                            end
                        end
                    end
                    qqLon[:] = cmixRatioMini[:] + (0.5.*qqLon[:])
                    qqLat[:] = cmixRatioMini[:] + (0.5.*qqLat[:])
                    cqqLonFull[:,:,iLev,iTrc] = qqLon[:,:]
                    cqqLatFull[:,:,iLev,iTrc] = qqLat[:,:]
                    # --- END CALC_ADVEC_CROSS_TERMS ---
                end
            end
        end
        for iWorker in workers()]
    end

    #DEBUG
    #newTime = toc()
    #@printf("Checkpoint Alfa - %10.4e\n",newTime - lastTime)
    #lastTime = 0.0
    #tic()

    # Had to synchronize as workers will be communicating during XADV and YADV calculation
    # Crude, crude, crude...
    xLonFull = dfill(0.0,(nLon,nLat,nLev,nTrc),workers(),[gridData.chunk3D...,1])
    xLatFull = dfill(0.0,(nLon,nLat,nLev,nTrc),workers(),[gridData.chunk3D...,1])
    @sync begin
        # 1: SP
        # 2: NP
        capSum = zeros(1,2,nLev,nTrc)
        [capSum += @fetchfrom iWorker begin
            localIdx = myindexes(mixRatioStart)
            localLon = localIdx[1]
            nLonLocal = length(localLon)
            localLat = localIdx[2]
            nLatLocal = length(localLat)
            cmixRatio = localpart(mixRatioStart)
            cdpStart = localpart(dpStart)
            cUCA = localpart(UCAvg)
            cVCA = localpart(VCAvg)
            nTrc = size(cmixRatio,4)
            nLev = size(cmixRatio,3)
            nLon = remotecall_fetch(myid(),()->nLon)
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            latRange = remotecall_fetch(myid(),()->latRangeSub)
            cqqLonFull = localpart(qqLonFull)
            cqqLatFull = localpart(qqLatFull)
            cxLonFull = localpart(xLonFull)
            cxLatFull = localpart(xLatFull)
            # Each core has a full complement of levels and tracers
            miniCap = zeros(1,2,nLev,nTrc)
            for iTrc in 1:nTrc
                for iLev in 1:nLev
                    minValidS = latRange[1,iLev]
                    maxValidN = latRange[2,iLev]
                    cqqLon = cqqLonFull[:,:,iLev,iTrc]
                    cqqLat = cqqLatFull[:,:,iLev,iTrc]
                    xLon = zeros(nLonLocal,nLatLocal)
                    xLat = zeros(nLonLocal,nLatLocal)
                    # Cap sum
                    # Only 2nd-order implemented (hard-coded in GC)
                    for iLat = 1:nLatLocal
                        # Only bother with non-polar-cap
                        targLat = localLat[iLat]
                        if targLat > nCap && targLat <= (nLat - nCap)
                            # TPCORE has an if-else as in the next (commented)
                            # line, but both outcomes appear to be identical
                            #if targLat < minValidS || targLat > maxValidN
                                # Different to above!
                                for iLon = 1:nLonLocal
                                    avgCourant = cUCA[iLon,iLat,iLev]
                                    iCFL = int(avgCourant)
                                    rCFL = iCFL - avgCourant
                                    tLon = iLon - iCFL
                                    # Need to check this three times now...
                                    qVals = zeros(3)
                                    for iVal = [-1 0 1]
                                        tLonIn = tLon + iVal
                                        if tLonIn < 1 || tLonIn > nLonLocal
                                            # Communicate
                                            tLonOut = 1 + mod(iVal - 1 + localLon[iLon] - iCFL,nLon)
                                            #if tLonOut > size(qqLatFull,1) || targLat > size(qqLatFull,2)
                                            #    #?!?!
                                            #    @printf("tLO iV lLi iL iC nL: %i %i %i %i %i %i\n",tLonOut,iVal,lonLocal[iLon],iLon,iCFL,nLon)
                                            #else
                                                qVals[iVal+2] = qqLatFull[tLonOut,targLat,iLev,iTrc]
                                            #end
                                        else
                                            # Local
                                            qVals[iVal+2] = cqqLat[tLonIn,iLat]
                                        end
                                    end
                                    aTemp = 0.5.*((qVals[3] + qVals[1])) - qVals[2]
                                    bTemp = 0.5.*(qVals[3] - qVals[1])
                                    cTemp = qVals[2] - cqqLat[iLon,iLat]
                                    xLon[iLon,iLat] = (rCFL*((aTemp*rCFL)+bTemp)) + cTemp
                                end
                            #else
                                # Non-polar area
                            #end
                        else
                            # Polar xLon = 0
                            xLon[:,iLat] = 0.0
                        end
                        # Now the latitudinal cross advection
                        # Note that here we impinge on the polar caps
                        if targLat >= nCap && targLat <= (nLat+1-nCap)
                            for iLon = 1:nLonLocal
                                avgCourant = cVCA[iLon,iLat,iLev]
                                iCFL = int(avgCourant)
                                rCFL = iCFL - avgCourant
                                tLat = iLat - iCFL
                                qVals = zeros(3)
                                targLon = localLon[iLon]
                                for iVal = [-1 0 1]
                                    tLatIn = tLat + iVal
                                    if tLatIn < 1 || tLatIn > nLatLocal
                                        # Communicate
                                        tLatOut = localLat[iLat] + iVal - iCFL
                                        qVals[iVal+2] = qqLonFull[targLon,tLatOut,iLev,iTrc]
                                    else
                                        qVals[iVal+2] = cqqLon[iLon,tLatIn]
                                    end
                                end
                                aTemp = 0.5.*((qVals[3] + qVals[1])) - qVals[2]
                                bTemp = 0.5.*(qVals[3] - qVals[1])
                                cTemp = qVals[2] - cqqLon[iLon,iLat]
                                xLat[iLon,iLat] = (rCFL*((aTemp*rCFL)+bTemp)) + cTemp
                                if targLat == nCap
                                    # Add to cap sum 1
                                    miniCap[1,1,iLev,iTrc] += xLat[iLon,iLat]
                                elseif targLat == (nLat + 1 - nCap)
                                    # Add to cap sum 2
                                    miniCap[1,2,iLev,iTrc] += xLat[iLon,iLat]
                                end
                            end
                        end
                    end
                    # Store xLon and xLat
                    cxLonFull[:,:,iLev,iTrc] = xLon[:,:]
                    cxLatFull[:,:,iLev,iTrc] = xLat[:,:]
                end # 1:nLev
            end # 1:nTrc
            # Return miniCap for capSum
            miniCap
        end
        for iWorker in workers()]
    end # SYNC

    #DEBUG
    #newTime = toc()
    #@printf("Checkpoint Bravo - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()

    # Polar averaging
    capSum = capSum./nLon
    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(mixRatioStart)
            nLonLocal = length(localIdx[1])
            localLat = localIdx[2]
            nLatLocal = length(localLat)
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            # Each core has a full complement of levels and tracers
            cxLatFull = localpart(xLatFull)
            cxLonFull = localpart(xLonFull)
            miniCap = remotecall_fetch(myid(),()->capSum)
            cmixRatio = localpart(mixRatioStart) # In original, update is applied to mixRatio... be careful
            for iLat = 1:nLatLocal
                targLat = localLat[iLat]
                if targLat <= nCap
                    for iLon = 1:nLonLocal
                        cxLatFull[iLon,iLat,:,:] = miniCap[1,1,:,:]
                    end
                elseif targLat > (nLat-nCap)
                    for iLon = 1:nLonLocal
                        cxLatFull[iLon,iLat,:,:] = miniCap[1,2,:,:]
                    end
                end
                # Perform update on mixRatio array
                cmixRatio[:,iLat,:,:] += cxLonFull [:,iLat,:,:] + cxLatFull[:,iLat,:,:]
            end
        end
        for iWorker in workers()]
    end

    #DEBUG
    #newTime = toc()
    #@printf("Checkpoint Charlie - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()

    # ---START XMIST---
    # Need to calculate horizontal slopes in qqLat
    if lonOrd == 1
        dcLon = dfill(0.0,nworkers(),workers())
    else
        dcLon = dfill(0.0,(nLon,nLat,nLev,nTrc),workers(),[gridData.chunk3D...,1])
        @sync begin
            [@spawnat iWorker begin
                localIdx = myindexes(mixRatioStart)
                localLon = localIdx[1]
                nLonLocal = length(localLon)
                localLat = localIdx[2]
                nLatLocal = length(localLat)
                cdcLon = localpart(dcLon)
                cmixRatio = localpart(mixRatioStart)
                nTrc = size(cmixRatio,4)
                nLev = size(cmixRatio,3)
                nLon = remotecall_fetch(myid(),()->nLon)
                nLat = remotecall_fetch(myid(),()->nLat)
                nCap = remotecall_fetch(myid(),()->nCap)
                cqqLatFull = localpart(qqLatFull)
                # Each core has a full complement of levels and tracers
                r24 = 1.0/24.0
                for iTrc in 1:nTrc
                    for iLev in 1:nLev
                        cqqLatMini = cqqLatFull[:,:,iLev,iTrc]
                        for iLat = 1:nLatLocal
                            targLat = localLat[iLat]
                            if targLat > (nCap+1) && targLat < (nLat - nCap)
                                for iLon = 1:nLonLocal
                                    targLon = localLon[iLon]
                                    qVals = zeros(5)
                                    for iVal in -2:2
                                        tLon = iLon + iVal
                                        if tLon < 1 || tLon > nLonLocal
                                            # Communicate
                                            tLonOut = targLon + iVal
                                            qVals[iVal+3] = qqLatFull[tLonOut,targLat,iLev,iTrc]
                                        else
                                            qVals[iVal+3] = cqqLatMini[tLon,iLat]
                                        end
                                    end
                                    tempVal = ((8.0*(qVals[4] - qVals[2])) + (qVals[1] - qVals[5])) * r24
                                    pMax = maximum(qVals[2:4])-qVals[3]
                                    pMin = qVals[3] - minimum(qVals[2:4])
                                    cdcLon[iLon,iLat,iLev,iTrc] = abs(min(abs(tempVal),pMax,pMin))*sign(tempVal)
                                end
                            end
                        end
                    end # nLev
                end # nTrc
            end
            for iWorker in workers()]
        end
    end

    # ---END XMIST---
    #DEBUG
    #newTime = toc()
    #@printf("Checkpoint Delta - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()

    latOffset = max(1,itrunc(nLat/18))
    fxLon = dfill(0.0,(nLon,nLat,nLev,nTrc),workers(),[gridData.chunk3D...,1])

    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(mixRatioStart)
            localLon = localIdx[1]
            nLonLocal = length(localLon)
            localLat = localIdx[2]
            nLatLocal = length(localLat)
            cdcLon = localpart(dcLon)
            cmixRatio = localpart(mixRatioNew)
            cdpStart = localpart(dpStart)
            nTrc = size(cmixRatio,4)
            nLev = size(cmixRatio,3)
            nLon = remotecall_fetch(myid(),()->nLon)
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            cqqLatFull = localpart(qqLatFull)
            cqqLonFull = localpart(qqLonFull)
            latRange = remotecall_fetch(myid(),()->latRangeSub)
            clonOrd = remotecall_fetch(myid(),()->lonOrd)
            clonLim = remotecall_fetch(myid(),()->lonLim)
            if clonLim <= 2
                # Load the LMTPPM function
                include("JZSpcFnc.jl")
            end
            clatOff = remotecall_fetch(myid(),()->latOffset)
            cUC = localpart(UCourant)
            clonFlux = localpart(lonFlux)
            cdpShift = localpart(dpShift)
            ratio23 = 2.0/3.0
            #showDebug = iWorker == 2
            showDebug = false
            # Will need these later
            ratio13 = 1.0/3.0
            a6 = zeros(nLonLocal + 2)
            al = zeros(nLonLocal + 2)
            ar = zeros(nLonLocal + 2)
            qa = zeros(nLonLocal + 2)
            dc = zeros(nLonLocal + 2)
            # Each core has a full complement of levels and tracers
            for iTrc in 1:nTrc
                # DEBUG
                if showDebug
                    @printf("Worker %i starting on tracer %i.\n",iWorker,iTrc)
                    tStart = time_ns()
                    nStagePoints = 10
                    stageCount = zeros(nStagePoints)
                    stageSum = zeros(nStagePoints)
                end
                for iLev in 1:nLev
                    minValidS = latRange[1,iLev]
                    maxValidN = latRange[2,iLev]
                    fLon = zeros(nLonLocal,nLatLocal)
                    cqqLatMini = cqqLatFull[:,:,iLev,iTrc]
                    cdcLonMini = cdcLon[:,:,iLev,iTrc]
                    for iLat = 1:nLatLocal
                        targLat = localLat[iLat]
                        if targLat >= minValidS && targLat <= maxValidN
                            if clonOrd == 1 || (targLat == (1+nCap)) || (targLat == (nLat-nCap))
                                # DEBUG
                                if showDebug && iLev == 1
                                    iStage = 1
                                    tNow = time_ns()
                                    tStage = tNow - tStart
                                    stageCount[iStage] += 1
                                    stageSum[iStage] += tStage
                                    tStart = tNow
                                end
                                for iLon = 1:nLonLocal
                                    targLon = localLon[iLon]
                                    fCFL = cUC[iLon,iLat,iLev]
                                    tLon = itrunc(iLon - fCFL)
                                    if tLon < 1 || tLon > nLonLocal
                                        # Communicate
                                        tLonOut = 1 + mod(itrunc(targLon-fCFL)-1,nLon)
                                        fLon[iLon,iLat] = qqLatFull[tLonOut,targLat,iLev,iTrc]
                                    else
                                        fLon[iLon,iLat] = cqqLatMini[tLon,iLat]
                                    end
                                end
                            else
                                # DEBUG
                                if showDebug && iLev == 1
                                    iStage = 2
                                    tNow = time_ns()
                                    tStage = tNow - tStart
                                    stageCount[iStage] += 1
                                    stageSum[iStage] += tStage
                                    tStart = tNow
                                end
                                if (clonOrd == 2) || (targLat <= (1+clatOff)) || (targLat >= (nLat-clatOff))
                                    # DEBUG
                                    if showDebug && iLev == 1
                                        iStage = 3
                                        tNow = time_ns()
                                        tStage = tNow - tStart
                                        stageCount[iStage] += 1
                                        stageSum[iStage] += tStage
                                        tStart = tNow
                                    end
                                    for iLon = 1:nLonLocal
                                        fCFL = cUC[iLon,iLat,iLev]
                                        targLon = localLon[iLon]
                                        tLon = itrunc(iLon - fCFL)
                                        if tLon < 1 || tLon > nLonLocal
                                            # Communicate
                                            tLonOut = 1 + mod(itrunc(targLon-fCFL)-1,nLon)
                                            qVal = qqLatFull[tLonOut,targLat,iLev,iTrc]
                                            dcVal = dcLon[tLonOut,targLat,iLev,iTrc]
                                        else
                                            qVal = cqqLatMini[tLon,iLat]
                                            dcVal = cdcLonMini[tLon,iLat]
                                        end
                                        fLon[iLon,iLat] = qVal + (dcVal*(sign(fCFL) - fCFL))
                                    end
                                else
                                    # DEBUG
                                    if showDebug && iLev == 1
                                        iStage = 4
                                        tNow = time_ns()
                                        tStage = tNow - tStart
                                        stageCount[iStage] += 1
                                        stageSum[iStage] += tStage
                                        tStart = tNow
                                    end
                                    #Fxppm!
                                    # 1-D operations
                                    a6[:] = 0.0
                                    al[:] = 0.0
                                    ar[:] = 0.0
                                    qa[:] = 0.0
                                    dc[:] = 0.0
                                    # Fill out the qa and dc vectors
                                    qa[2:end-1] = cqqLatMini[:,iLat]
                                    dc[2:end-1] = cdcLonMini[:,iLat]
                                    # First index
                                    targLon = 1 + mod(localLon[1] - 2,nLon)
                                    qa[1] = qqLatFull[targLon,targLat,iLev,iTrc]
                                    dc[1] = dcLon[targLon,targLat,iLev,iTrc]
                                    # Last index
                                    targLon = 1 + mod(localLon[end],nLon)
                                    qa[end] = qqLatFull[targLon,targLat,iLev,iTrc]
                                    dc[end] = dcLon[targLon,targLat,iLev,iTrc]
                                    # DEBUG
                                    if showDebug && iLev == 1
                                        iStage = 5
                                        tNow = time_ns()
                                        tStage = tNow - tStart
                                        stageCount[iStage] += 1
                                        stageSum[iStage] += tStage
                                        tStart = tNow
                                    end
                                    for iLon = 2:nLonLocal+2
                                        rVal = 0.5.*(sum(qa[iLon-1:iLon])) + (dc[iLon-1]-dc[iLon])*ratio13
                                        al[iLon] = rVal
                                        ar[iLon-1] = rVal
                                    end
                                    for iLon = 1:nLonLocal+2
                                        a6[iLon] = 3.0*((2.0*qa[iLon]) - (al[iLon]+ar[iLon]))
                                    end
                                    # DEBUG
                                    if showDebug && iLev == 1
                                        iStage = 6
                                        tNow = time_ns()
                                        tStage = tNow - tStart
                                        stageCount[iStage] += 1
                                        stageSum[iStage] += tStage
                                        tStart = tNow
                                    end
                                    if clonLim <= 2
                                        (a6,al,ar) = Lmtppm(clonLim,a6,al,ar,dc,qa)
                                        for iLon = 1:nLonLocal
                                            fCFL = cUC[iLon,iLat,iLev]
                                            if fCFL > 0.0
                                                fLon[iLon,iLat] = ar[iLon] + 0.5*fCFL*(al[iLon] - ar[iLon] + 
                                                    (a6[iLon] * (1.0 - (ratio23 * fCFL))))
                                            else
                                                fLon[iLon,iLat] = al[iLon+1] - 0.5*fCFL*(ar[iLon+1] - al[iLon+1] + 
                                                    (a6[iLon+1] * (1.0 + (ratio23 * fCFL))))
                                            end
                                        end
                                    end
                                    # DEBUG
                                    if showDebug && iLev == 1
                                        iStage = 7
                                        tNow = time_ns()
                                        tStage = tNow - tStart
                                        stageCount[iStage] += 1
                                        stageSum[iStage] += tStage
                                        tStart = tNow
                                    end
                                end
                            end
                            fLon[:,iLat] = fLon[:,iLat] .* clonFlux[:,iLat,iLev]
                        else
                            # Horizonal conservative (flux-form) semi-Lagrangian advection
                            # van Leer at high altitudes
                            # DEBUG
                            if showDebug && iLev == 1
                                iStage = 8
                                tNow = time_ns()
                                tStage = tNow - tStart
                                stageCount[iStage] += 1
                                stageSum[iStage] += tStage
                                tStart = tNow
                            end
                            if (clonOrd == 1) || (iLat == (1+nCap)) || (iLat == (nLat-nCap))
                                for iLon = 1:nLonLocal
                                    targLon = localLon[iLon]
                                    fCFL = cUC[iLon,iLat,iLev]
                                    tLon = itrunc(iLon - fCFL)
                                    rCFL = fCFL - itrunc(fCFL)
                                    if tLon < 1 || tLon > nLonLocal
                                        # Communicate
                                        tLonOut = 1 + mod(itrunc(targLon-fCFL)-1,nLon)
                                        qVal = qqLatFull[tLonOut,targLat,iLev,iTrc]
                                    else
                                        qVal = cqqLatMini[tLon,iLat]
                                    end
                                    fLon[iLon,iLat] = rCFL * qVal
                                end
                            else
                                for iLon = 1:nLonLocal
                                    targLon = localLon[iLon]
                                    fCFL = cUC[iLon,iLat,iLev]
                                    # isav[iLon] = iLon - itrunc(fCFL)
                                    tLon = itrunc(iLon - fCFL)
                                    rCFL = fCFL - itrunc(fCFL)
                                    if tLon < 1 || tLon > nLonLocal
                                        # Communicate
                                        tLonOut = 1 + mod(itrunc(targLon-fCFL)-1,nLon)
                                        qVal = qqLatFull[tLonOut,targLat,iLev,iTrc]
                                        dcVal = dcLon[tLonOut,targLat,iLev,iTrc]
                                    else
                                        qVal = cqqLatMini[tLon,iLat]
                                        dcVal = cdcLonMini[tLon,iLat]
                                    end
                                    fLon[iLon,iLat] = qVal + (dcVal * (sign(rCFL) - rCFL))
                                end
                            end
                            # DEBUG
                            if showDebug && iLev == 1
                                iStage = 9
                                tNow = time_ns()
                                tStage = tNow - tStart
                                stageCount[iStage] += 1
                                stageSum[iStage] += tStage
                                tStart = tNow
                            end
                            # TPCORE has isav here, but we don't bother
                            # Deal with C>1
                            for iLon = 1:nLonLocal
                                fCFL = cUC[iLon,iLat,iLev]
                                tSav = iLon - itrunc(fCFL)
                                targLon = localLon[iLon]
                                tSavOut = 1 + mod(targLon - (1+itrunc(fCFL)),nLon)
                                if fCFL > 1.0
                                    for tLon = tSav:(iLon-1)
                                        if tLon < 1 || tLon > nLonLocal
                                            tLonOut = 1 + mod(localLon[iLon] - 1 + tLon - 1,nLon)
                                            qVal = qqLatFull[tLonOut,targLat,iLev,iTrc]
                                        else
                                            qVal = cqqLatMini[tLon,iLat]
                                        end
                                        fLon[iLon,iLat] += qVal
                                    end
                                elseif fCFL < -1.0
                                    for tLon = iLon:(tSav-1)
                                        if tLon < 1 || tLon > nLonLocal
                                            tLonOut = 1 + mod(localLon[iLon] - 1 + tLon - 1,nLon)
                                            qVal = qqLatFull[tLonOut,targLat,iLev,iTrc]
                                        else
                                            qVal = cqqLatMini[tLon,iLat]
                                        end
                                        fLon[iLon,iLat] -= qVal
                                    end
                                end
                            end
                            fLon[:,iLat] = fLon[:,iLat] .* cdpShift[:,iLat,iLev]
                            # DEBUG
                            if showDebug && iLev == 1
                                iStage = 10
                                tNow = time_ns()
                                tStage = tNow - tStart
                                stageCount[iStage] += 1
                                stageSum[iStage] += tStage
                                tStart = tNow
                            end
                        end
                    end
                    # Update outer arrays
                    cfxLon = localpart(fxLon)
                    cfxLon[:,:,iLev,iTrc] = fLon[:]
                    if showDebug && iLev == 1
                        @printf("Local surface grid processed (%i grid points)\n",nLonLocal * nLatLocal)
                        for iStage = 1:nStagePoints
                            @printf("Point #%02i time: %12.6fs [%03i samples]\n",iStage,(stageSum[iStage]*1.0e-9)./stageCount[iStage],stageCount[iStage])
                        end
                        stageSum[:] = 0.0
                        stageCount[:] = 0
                    end
                end # nLev
            end # nTrc
        end
        for iWorker in workers()]
    end

    # DEBUG
    #newTime = toc()
    #@printf("Checkpoint Echo - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()

    # Final loop of Xtp

    # Needed data to be fully available before updating this...
    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(mixRatioStart)
            localLon = localIdx[1]
            nLonLocal = length(localLon)
            localLat = localIdx[2]
            nLatLocal = length(localLat)
            nTrc = length(localIdx[4])
            nLev = length(localIdx[3])
            nLon = remotecall_fetch(myid(),()->nLon)
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            cdMRStart = localpart(dMRStartFull)
            fLon = localpart(fxLon)
            # Each core has a full complement of levels and tracers
            for iLat = 1:nLatLocal
                targLat = localLat[iLat]
                if targLat > nCap && targLat <= (nLat - nCap)
                    for iLev = 1:nLev
                        for iTrc = 1:nTrc
                            for iLon = 1:nLonLocal
                                targLon = localLon[iLon]
                                tLon = iLon + 1
                                if iLon == nLonLocal
                                    # Need to communicate
                                    tLonOut = targLon + 1
                                    if targLon == nLon
                                        tLonOut = 1
                                    end
                                    fLonP1 = fxLon[tLonOut,targLat,iLev,iTrc]
                                else
                                    fLonP1 = fLon[tLon,iLat,iLev,iTrc]
                                end
                                cdMRStart[iLon,iLat,iLev,iTrc] = cdMRStart[iLon,iLat,iLev,iTrc] + (fLon[iLon,iLat,iLev,iTrc] - fLonP1)
                            end
                        end
                    end
                end
            end
        end
        for iWorker in workers()]
    end

    # ---END XTP---

    # DEBUG
    #newTime = toc()
    #@printf("Checkpoint Foxtrot - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()

    # ---START YTP---
    dcLat = dfill(0.0,(nLon,nLat,nLev,nTrc),workers(),[gridData.chunk3D...,1])
    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(mixRatioStart)
            localLon = localIdx[1]
            nLonLocal = length(localLon)
            localLat = localIdx[2]
            nLatLocal = length(localLat)
            cdcLat = localpart(dcLat)
            cmixRatio = localpart(mixRatioStart)
            nTrc = size(cmixRatio,4)
            nLev = size(cmixRatio,3)
            nLon = remotecall_fetch(myid(),()->nLon)
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            cqqLonFull = localpart(qqLonFull)
            cqqLatFull = localpart(qqLatFull)
            clatOrd = remotecall_fetch(myid(),()->latOrd)
            cVC = localpart(VCourant)
            clatFlux = localpart(latFlux)
            ratio24 = 1.0/24.0
            for iTrc = 1:nTrc
                for iLev = 1:nLev
                    if clatOrd == 1
                        for iLat = 1:nLatLocal
                            targLat = localLat[iLat]
                            if targLat > nCap && targLat <= (nLat - nCap)
                                for iLon = 1:nLonLocal
                                    targLon = localLon[iLon]
                                    tLatOut = itrunc(1 + nCap - cVC[iLon,iLat,iLev])
                                    tLat = tLatOut + 1 - localLat[1]
                                    if tLat < 1 || tLat > nLatLocal
                                        qVal = qqLonFull[targLon,tLatOut,iLev,iTrc]
                                    else
                                        qVal = cqqLonFull[iLon,tLat,iLev,iTrc]
                                    end
                                    cqqLatFull[iLon,iLat,iLev,iTrc] = qVal
                                end
                            end # Non-cap
                        end # nLat
                    else
                        # ---START YMIST---
                        # Short on time - ignoring unused types
                        for iLat = 1:nLatLocal
                            for iLon = 1:nLonLocal
                                targLon = localLon[iLon]
                                qVals = zeros(5)
                                for iVal = [-2 -1 0 1 2]
                                    tLat = iLat + iVal
                                    if tLat < 1 || tLat > nLatLocal
                                        # Communication required
                                        tLatOut = localLat[iLat] + iVal
                                        # Ignore non-physical values (leave as 0)
                                        if tLatOut > 0 && tLatOut <= nLat
                                            qVals[iVal+3] = qqLonFull[targLon,tLatOut,iLev,iTrc]
                                        end
                                    else
                                        # Internal
                                        qVals[iVal+3] = cqqLonFull[iLon,tLat,iLev,iTrc]
                                    end
                                end
                                tempVal = ((8.0*(qVals[4]-qVals[2])) + (qVals[1]-qVals[5])) * ratio24
                                pMax = maximum(qVals[2:4]) - qVals[3]
                                pMin = qVals[3] - minimum(qVals[2:4])
                                cdcLat[iLon,iLat,iLev,iTrc] = min(abs(tempVal),pMin,pMax)*sign(tempVal)
                            end
                        end
                        # ---END YMIST---
                    end # if clatOrd
                end # nLev
            end # nTrc
        end
        for iWorker in workers()]
    end

    #newTime = toc()
    #@printf("Checkpoint Golf - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()
    fxLat = dfill(0.0,(nLon,nLat,nLev,nTrc),workers(),[gridData.chunk3D...,1])
    geoFactor = gridData.geoFactor
    capFactor = gridData.capFactor

    # Only necessary if cLatOrd != 1
    if latOrd != 1
        @sync begin
            [@spawnat iWorker begin
                localIdx = myindexes(mixRatioStart)
                localLon = localIdx[1]
                nLonLocal = length(localLon)
                localLat = localIdx[2]
                nLatLocal = length(localLat)
                cdcLat = localpart(dcLat)
                cmixRatio = localpart(mixRatioStart)
                cdpStart = localpart(dpStart)
                nTrc = size(cmixRatio,4)
                nLev = size(cmixRatio,3)
                nLon = remotecall_fetch(myid(),()->nLon)
                nLat = remotecall_fetch(myid(),()->nLat)
                nCap = remotecall_fetch(myid(),()->nCap)
                cqqLonFull = localpart(qqLonFull)
                cqqLatFull = localpart(qqLatFull)
                latRange = remotecall_fetch(myid(),()->latRangeSub)
                clatOrd = remotecall_fetch(myid(),()->latOrd)
                clatLim = remotecall_fetch(myid(),()->latLim)
                if clatLim <= 2
                    # Load the LMTPPM function
                    include("JZSpcFnc.jl")
                end
                cVC = localpart(VCourant)
                clatFlux = localpart(latFlux)
                cdpShift = localpart(dpShift)
                ratio24 = 1.0/24.0
                for iLev = 1:nLev
                    for iTrc = 1:nTrc
                        cqqLonMini = cqqLonFull[:,:,iLev,iTrc]
                        cdcLatMini = cdcLat[:,:,iLev,iTrc]
                        if clatOrd <= 0 || clatOrd >= 3
                            #Fyppm!
                            a6 = zeros(nLonLocal,nLatLocal+2)
                            al = zeros(nLonLocal,nLatLocal+2)
                            ar = zeros(nLonLocal,nLatLocal+2)
                            dc = zeros(nLonLocal,nLatLocal+2)
                            qa = zeros(nLonLocal,nLatLocal+2)
                            ratio13 = 1.0/3.0
                            ratio23 = 2.0/3.0
                            #------------------
                            # CORDON
                            #------------------
                            # Fill out the qa and dc vectors
                            qa[:,2:end-1] = cqqLonMini[:,:]
                            dc[:,2:end-1] = cdcLatMini[:,:]
                            # First index
                            targLat = localLat[1] - 1
                            if targLat > 0
                                qa[:,1] = qqLonFull[localLon,targLat,iLev,iTrc]
                                dc[:,1] = dcLat[localLon,targLat,iLev,iTrc]
                            end
                            # Last index
                            targLat = 1 + localLat[end]
                            if targLat <= nLat
                                qa[:,end] = qqLonFull[localLon,targLat,iLev,iTrc]
                                dc[:,end] = dcLat[localLon,targLat,iLev,iTrc]
                            end
                            for iLat = 2:nLatLocal+2
                                rVal = 0.5.*(sum(qa[:,iLat-1:iLat],2) + (dc[:,iLat-1]-dc[:,iLat])).*ratio13
                                al[:,iLat] = rVal
                                ar[:,iLat-1] = rVal
                            end
                            #------------------
                            # CORDON
                            #------------------
                           for iLat = 1:nLatLocal+2
                                a6[:,iLat] = 3.0.*((2.0.*qa[:,iLat]) - (al[:,iLat]+ar[:,iLat]))
                            end
                            if clatLim <= 2
                                # Load the LMTPPM function
                                include("JZSpcFnc.jl")
                                for iLat = 1:nLatLocal+2
                                    (a6[:,iLat],al[:,iLat],ar[:,iLat]) = Lmtppm(clatLim,a6[:,iLat],al[:,iLat],ar[:,iLat],dc[:,iLat],qa[:,iLat])
                                end
                            end
                            for iLat = 1:nLatLocal
                                targLat = localLat[iLat]
                                for iLon = 1:nLonLocal
                                    targLon = localLon[iLon]
                                    fCFL = cVC[iLon,iLat,iLev]
                                    if fCFL > 0.0
                                        cqqLatFull[iLon,iLat,iLev,iTrc] = ar[iLon,iLat] + 0.5.*fCFL.*(al[iLon,iLat]-ar[iLon,iLat]+(a6[iLon,iLat].*(1.0-(ratio23.*fCFL))))
                                    else
                                        cqqLatFull[iLon,iLat,iLev,iTrc] = al[iLon,iLat+1] - 0.5.*fCFL.*(ar[iLon,iLat+1]-al[iLon,iLat+1]+(a6[iLon,iLat+1].*(1.0+(ratio23.*fCFL))))
                                    end
                                end
                            end
                        else
                            for iLat = 1:nLatLocal
                                targLat = localLat[iLat]
                                if targLat > nCap && targLat <= (nLat + 1 - nCap)
                                    for iLon = 1:nLonLocal
                                        targLon = localLon[iLon]
                                        tLatOut = itrunc(1 + nCap - cVC[iLon,iLat,iLev])
                                        tLat = tLatOut + 1 - localLat[1]
                                        if tLat < 1 || tLat > nLatLocal
                                            qVal = qqLonFull[targLon,tLatOut,iLev,iTrc]
                                            dcVal = dcLat[targLon,tLatOut,iLev,iTrc]
                                        else
                                            qVal = cqqLonFull[iLon,tLat,iLev,iTrc]
                                            dcVal = cdcLat[iLon,tLat,iLev,iTrc]
                                        end
                                        cqqLatFull[iLon,iLat,iLev,iTrc] = qVal + (sign(cVC[iLon,iLat,iLev]) * dcVal)
                                    end
                                end
                            end
                        end
                    end #nLev
                end #nTrc
                # Perform local updates
                for iTrc in 1:nTrc
                    for iLat in 1:nLatLocal
                        for iLon in 1:nLonLocal
                            for iLev in 1:nLev
                                cqqLatFull[iLon,iLat,iLev,iTrc] = cqqLatFull[iLon,iLat,iLev,iTrc]*clatFlux[iLon,iLat,iLev]
                            end
                        end
                    end
#                    cqqLatFull[:,:,:,iTrc] = cqqLatFull[:,:,:,iTrc].*clatFlux[:,:,:]
                end
                fLat = localpart(fxLat)
                cgeoFac = remotecall_fetch(myid(),()->geoFactor)
                for iLat = 1:nLatLocal
                    targLat = localLat[iLat]
                    fLat[:,iLat,:,:] = cqqLatFull[:,iLat,:,:] .* cgeoFac[targLat]
                end
            end
            for iWorker in workers()]
        end
    end

    #newTime = toc()
    #@printf("Checkpoint Hotel - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()
    
    @sync begin
        # Take this opportunity to retrieve polar cap data
        capSum = zeros(1,2,nLev,nTrc)
        [capSum += @fetchfrom iWorker begin
            localIdx = myindexes(mixRatioStart)
            localLon = localIdx[1]
            localLat = localIdx[2]
            nLatLocal = length(localLat)
            cdcLat = localpart(dcLat)
            cdMRStart = localpart(dMRStartFull)
            cgeoFac = remotecall_fetch(myid(),()->geoFactor)
            cqqLatFull = localpart(qqLatFull)
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            miniSum = zeros(1,2,nLev,nTrc)
            for iLat = 1:nLatLocal
                targLat = localLat[iLat]
                if targLat > nCap && targLat <= (nLat - nCap)
                    if iLat < nLatLocal
                        cdMRStart[:,iLat,:,:] = cdMRStart[:,iLat,:,:] + ((cqqLatFull[:,iLat,:,:] -
                            cqqLatFull[:,iLat+1,:,:]) .* cgeoFac[iLat])
                    else
                        # Communication required
                        cdMRStart[:,iLat,:,:] = cdMRStart[:,iLat,:,:] + ((cqqLatFull[:,iLat,:,:] -
                            convert(Array,qqLatFull[targLon,targLat+1,:,:])).*cgeoFac[iLat])
                    end
                    if targLat == (nCap + 1)
                        # Collect South Pole data
                        miniSum[1,1,:,:] += sum(cqqLatFull[:,iLat,:,:],1)
                    end
                elseif targLat == (nLat + 1 - nCap)
                    # Collect North Pole data
                    miniSum[1,2,:,:] += sum(cqqLatFull[:,iLat,:,:],1)
                end
            end
            miniSum
        end
        for iWorker in workers()]
    end

    # Minor local processing
    capSum = capSum ./ (nLon * capFactor)
    capFetch = RemoteRef(myid())
    put(capFetch,capSum)

    #newTime = toc()
    #@printf("Checkpoint India - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()

    # Calculate polar cap
    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(mixRatioStart)
            localLon = localIdx[1]
            localLat = localIdx[2]
            nLatLocal = length(localLat)
            nLonLocal = length(localLon)
            cdcLat = localpart(dcLat)
            cdMRStart = localpart(dMRStartFull)
            cqqLatFull = localpart(qqLatFull)
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            # Do we need to do anything?
            hasSP = any(localLat .<= nCap)
            hasNP = any(localLat .>= (nLat + 1 - nCap))
            if hasSP || hasNP
                miniSum = fetch(capFetch)
                miniSumOff = copy(miniSum)
                if hasSP
                    miniSumOff[1,1,:,:] = convert(Array,dMRStartFull[1,1,:,:]) - miniSum[1,1,:,:]
                end
                if hasNP
                    miniSumOff[1,2,:,:] = convert(Array,dMRStartFull[1,nLat,:,:]) + miniSum[1,2,:,:]
                end
                for iLat = 1:nLatLocal
                    targLat = localLat[iLat]
                    if targLat < (nCap + 1)
                        # Set to SP value
                        for iLon = 1:nLonLocal
                            cdMRStart[iLon,iLat,:,:] = miniSumOff[1,1,:,:]
                        end
                    elseif targLat >= (nLat + 1 - nCap)
                        # Set to NP value
                        for iLon = 1:nLonLocal
                            cdMRStart[iLon,iLat,:,:] = miniSumOff[1,2,:,:]
                        end
                    end
                end
            end
        end
        for iWorker in workers()]
    end

    # NOTE: fLat[:,1:2,:,:] and fLat[:,end:end+1] were NOT calculated

    # Apply fluxes
    
    #newTime = toc()
    #@printf("Checkpoint Juliet - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()


    # ---END YTP---

    # All levels were done for X & Y transport

    # ---START FZPPM---
    # Establish dpi equivalent array
    dpiArr = dfill(0.0,(nLon,nLat,nLev,nTrc),workers(),[gridData.chunk3D...,1])
    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(mixRatioStart)
            cMR = localpart(mixRatioStart)
            nLev = length(localIdx[3])
            cdpiArr = localpart(dpiArr)
            for iLev in nLev:-1:2
                cdpiArr[:,:,iLev,:] = cMR[:,:,iLev-1,:] - cMR[:,:,iLev,:]
            end
        end
        for iWorker in workers()]
    end

    #newTime = toc()
    #@printf("Checkpoint Kilo - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()

    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(mixRatioStart)
            localLon = localIdx[1]
            localLat = localIdx[2]
            nLatLocal = length(localLat)
            nLonLocal = length(localLon)
            nLev = length(localIdx[3])
            nTrc = length(localIdx[4])
            cdcLat = localpart(dcLat)
            cdMRStart = localpart(dMRStartFull)
            cqqLatFull = localpart(qqLatFull)
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            cdpiArr = localpart(dpiArr)
            cdpStart = localpart(dpStart)
            clevLim = remotecall_fetch(myid(),()->levLim)
            cvertMassFlux = localpart(vertMassFlux)
            ratio13 = 1.0/3.0
            ratio23 = 2.0/3.0
            cMR = localpart(mixRatioStart)
            for iTrc in 1:nTrc
#                dcLocal = zeros(nLonLocal,nLatLocal,nLevLocal)
                for iLat in 1:nLatLocal
                    targLat = localLat[iLat]
                    for iLon in 1:nLonLocal
                        cdpVec = cdpStart[iLon,iLat,:]
                        cMRVec = cMR[iLon,iLat,:,iTrc]
                        cdpiVec = cdpiArr[iLon,iLat,:,iTrc]
                        a6 = zeros(nLev)
                        al = zeros(nLev)
                        ar = zeros(nLev)
                        dc = zeros(nLev)
                        for iLev in nLev-1:-1:2
                            c0 = cdpVec[iLev]/sum(cdpVec[iLev-1:iLev+1])
                            c1 = (cdpVec[iLev+1]+(0.5*cdpVec[iLev]))/
                                 (cdpVec[iLev-1]+cdpVec[iLev])
                            c2 = (cdpVec[iLev-1]+(0.5*cdpVec[iLev]))/
                                 (cdpVec[iLev+1]+cdpVec[iLev])
                            cTmp = c0 * ((c1*cdpiVec[iLev]) + ((c2*cdpiVec[iLev+1])))
                            qMax = maximum(cMRVec[iLev-1:iLev+1]) - cMRVec[iLev]
                            qMin = cMRVec[iLev] - minimum(cMRVec[iLev-1:iLev+1])
#                            dcLocal[iLon,iLat,iLev] = min(abs(cTmp),qMax,qMin)*sign(cTmp)
                            dc[iLev] = min(abs(cTmp),qMax,qMin)*sign(cTmp)
                        end # nLev
                        if (targLat != nCap) && (targLat != (nLat + 1 - nCap))
                            # Model top...
                            fac1 =  cdpiVec[nLev-1] - cdpiVec[nLev] * (cdpVec[nLev-1]+cdpVec[nLev-2]) / 
                                    (cdpVec[nLev] + cdpVec[nLev-1])
                            fac2 =  (cdpVec[nLev-1] + cdpVec[nLev-2]) * (cdpVec[nLev] + cdpVec[nLev-1] + cdpVec[nLev-2])
                            aaFac = 3.0*fac1/fac2
                            bbFac = 2.0*cdpiVec[nLev]/(cdpVec[nLev]+cdpVec[nLev-1]) - 
                                    ratio23 * aaFac * (2.0 * cdpVec[nLev] + cdpVec[nLev-1])
                            al[nLev] = cMRVec[nLev] - cdpVec[nLev] * (ratio13*aaFac*cdpVec[nLev] + 0.5*bbFac)
                            al[nLev-1] = cdpVec[nLev]*(aaFac*cdpVec[nLev]+bbFac) + al[nLev]
                            if (cMRVec[nLev]*al[nLev] <= 0.0)
                                al[nLev] = 0.0
                                dc[nLev] = 0.0
                            else
                                dc[nLev] = cMRVec[nLev] - al[nLev]
                            end

                            # Model base...
                            fac1 = cdpiVec[2] * (cdpVec[1] * cdpVec[1]) / ((cdpVec[1]+cdpVec[2]) *
                                        (2.0*cdpVec[1] + cdpVec[2]))
                            ar[1] = cMRVec[1] + fac1
                            al[1] = cMRVec[1] - (fac1 + fac1)
                            if ((cMRVec[1]*ar[1])<=0.0)
                                ar[1] = 0.0
                            end
                            dc[1] = ar[1] - cMRVec[1]

                            # 4th-order interpolation within
                            for iLev in nLev-2:-1:2
                                c1 = (cdpiVec[iLev+1]*cdpVec[iLev+1]) / (cdpVec[iLev+1] + cdpVec[iLev])
                                c2 = 2.0/sum(cdpVec[iLev-1:iLev+2])
                                a1 = (cdpVec[iLev+2]+cdpVec[iLev+1]) / (2.0*cdpVec[iLev+1] + cdpVec[iLev])
                                a2 = (cdpVec[iLev]+cdpVec[iLev-1]) / (2.0*cdpVec[iLev] + cdpVec[iLev+1])
                                al[iLev] = cMRVec[iLev+1] + c1 + c2 * (cdpVec[iLev]*(c1*(a1-a2)+a2*dc[iLev+1]) - 
                                            cdpVec[iLev+1] * a1 * dc[iLev])
                            end # nLev

                            # Fill out what's left of ar
                            ar[2:end] = al[1:(end-1)]
                            a6[:] = 3.0*(2.0*cMRVec[:] - (al[:] + ar[:]))

                            # Force monotonicity in the bottom and top 2 layers
                            targRange = [1,2,nLev-1,nLev]
                            (a6New,alNew,arNew) = Lmtppm(0,a6[targRange],
                                             al[targRange],ar[targRange],
                                             dc[targRange],cMRVec[targRange])
                            al[targRange] = alNew[:]
                            ar[targRange] = arNew[:]
                            a6[targRange] = a6New[:]

                            # Interior depends on lonlmt
                            # In name of expediency, only bother with the case currently begin investigated
                            if clevLim == 4
                                for iLev in nLev-1 : -1 : 2
                                    dc[iLev] = cdpiVec[iLev] - cdpiVec[iLev+1]
                                end
                                for iLev in nLev-2 : -1 : 3
                                    # Right edges
                                    qTemp   = cMRVec[iLev] + (2.0*cdpiVec[iLev+1])
                                    lac     = cMRVec[iLev] + (1.5*dc[iLev+1]) + (0.5*cdpiVec[iLev+1])
                                    qMin = min(cMRVec[iLev],qTemp,lac)
                                    qMax = max(cMRVec[iLev],qTemp,lac)
                                    ar[iLev] = min(max(ar[iLev],qMin),qMax)
                                    # Left edges
                                    qTemp   = cMRVec[iLev] - (2.0*cdpiVec[iLev])
                                    lac     = cMRVec[iLev] + (1.5*dc[iLev-1]) - (0.5*cdpiVec[iLev])
                                    qMin = min(cMRVec[iLev],qTemp,lac)
                                    qMax = max(cMRVec[iLev],qTemp,lac)
                                    al[iLev] = min(max(al[iLev],qMin),qMax)
                                    # Recompute a6 accordingly
                                    a6[iLev] = 3.0 * (2.0*cMRVec[iLev] - (ar[iLev] + al[iLev]))
                                end
                            else
                                error("Code for lonOrd !=7 is not in place")
                            end
                            cwzVec = cvertMassFlux[iLon,iLat,:]
                            for iLev = nLev:-1:2
                                if cwzVec[iLev] > 0.0
                                    cm = cwzVec[iLev]/cdpVec[iLev]
                                    dc[iLev-1] = ar[iLev] + 0.5*cm*(al[iLev] - ar[iLev] + 
                                        a6[iLev] * (1.0 - ratio23 * cm))
                                else
                                    cp = cwzVec[iLev] / cdpVec[iLev-1]
                                    dc[iLev-1] = al[iLev-1] + 0.5*cp*(al[iLev-1]-ar[iLev-1]-
                                        a6[iLev-1] * (1.0 + ratio23 * cp))
                                end
                            end
                            for iLev = nLev:-1:2
                                dc[iLev-1] = cwzVec[iLev] * dc[iLev-1]
                            end
                            cdMR1Vec = cdMRStart[iLon,iLat,:,iTrc]
                            cdMR1Vec[nLev] -= dc[nLev-1]
                            cdMR1Vec[1] += dc[1]
                            for iLev = nLev-1:-1:2
                                cdMR1Vec[iLev] += (dc[iLev] - dc[iLev-1])
                            end
                            # Reassign for safety
                            cdMRStart[iLon,iLat,:,iTrc] = cdMR1Vec[:]
                        end # Cap check
                    end #nLon
                end # nLat
            end # nTrc
        end
        for iWorker in workers()]
    end
    # ---END FZPPM---

    # Synchronise
    #newTime = toc()
    #@printf("Checkpoint Lima - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()
    # Fill negative values
    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(mixRatioStart)
            localLon = localIdx[1]
            localLat = localIdx[2]
            nLatLocal = length(localLat)
            nLonLocal = length(localLon)
            cdcLat = localpart(dcLat)
            cdMRStart = localpart(dMRStartFull)
            cqqLatFull = localpart(qqLatFull)
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            cdpiArr = localpart(dpiArr)
            nLev = length(localIdx[3])
            clevLim = remotecall_fetch(myid(),()->levLim)
            cvertMassFlux = localpart(vertMassFlux)
            ratio13 = 1.0/3.0
            ratio23 = 2.0/3.0
            cMR = localpart(mixRatioStart)
            for iTrc in 1:nTrc
                for iLat in 1:nLatLocal
                    targLat = localLat[iLat]
                    if targLat > nCap || targLat <= (nLat - nCap)
                        for iLon in 1:nLonLocal
                            # Model top
                            cdMRVec = cdMRStart[iLon,iLat,:,iTrc]
                            if cdMRVec[nLev] < 0.0
                                cdMRVec[nLev-1] = cdMRVec[nLev-1] + cdMRVec[nLev]
                                cdMRVec[nLev] = 0.0
                            end
                            # Intermediate layers
                            for iLev = nLev-1:-1:2
                                if cdMRVec[iLev] < 0.0
                                    qUp = cdMRVec[iLev+1]
                                    qLy = -cdMRVec[iLev]
                                    dUp = min(qLy,qUp)
                                    cdMRVec[iLev+1] = qUp - dUp
                                    cdMRVec[iLev] = 0.0
                                    cdMRVec[iLev-1] = cdMRVec[iLev-1] + dUp - qLy
                                end
                            end
                            # Surface
                            if cdMRVec[1] < 0.0
                                qUp = cdMRVec[2]
                                qLy = -cdMRVec[1]
                                dUp = min(qLy,qUp)
                                cdMRVec[2] = qUp-dUp
                                cdMRVec[1] = 0.0
                            end
                            # Safety - transfer data to externally-communicating array
                            cdMRStart[iLon,iLat,:,iTrc] = cdMRVec[:]
                        end # nLon
                    end # Cap check
                end # nLat
            end # nTrc
        end
        for iWorker in workers()]
    end

    # Synchronise to make sure that dcLat is filled out and qqLat is up to date
    #newTime = toc()
    #@printf("Checkpoint Mike - %10.4e\n",newTime - lastTime)
    #lastTime = newTime
    #lastTime = 0.0
    #tic()

    # Copy the results over
    @sync begin
        [@spawnat iWorker begin
            cMR = localpart(mixRatioStart)
            #cUpdate = localpart(mixRatioNew)
            cdMRStart = localpart(dMRStartFull)
            cdpEnd = localpart(dpEnd)
            nCap = remotecall_fetch(myid(),()->nCap)
            nLat = remotecall_fetch(myid(),()->nLat)
            localIdx = myindexes(mixRatioNew)
            localLat = localIdx[2]
            localLon = localIdx[1]
            nLatLocal = length(localLat)
            nLonLocal = length(localLon)
            nTrc = length(localIdx[4])
            for iTrc = 1:nTrc
                for iLat = 1:nLatLocal
                    targLat = localLat[iLat]
                    isSP = targLat == nCap
                    isNP = targLat == (nLat + 1 - nCap)
                    if isSP || isNP
                        # South/North pole
                        if isSP
                            sourceLat = 1
                        else
                            sourceLat = nLat
                        end
                        findSource = localLat .== sourceLat
                        if any(findSource)
                            # Stored in the same worker
                            iLL = findfirst(findSource)
                            cMR[:,iLat,:,iTrc] = cdMRStart[:,iLL,:,iTrc]./cdpEnd[:,iLL,:]
                        else
                            # Communication req'd
                            cMR[:,iLat,:,iTrc] = dMRStartFull[localLon,sourceLat,:,iTrc]./dpEnd[localLon,sourceLat,:]
                        end
                    else
                        cMR[:,iLat,:,iTrc] = cdMRStart[:,iLat,:,iTrc]./cdpEnd[:,iLat,:]
                    end
                end
            end
        end
        for iWorker in workers()]
    end

    #newTime = toc()
    #@printf("Checkpoint November - %10.4e\n",newTime - lastTime)
    #lastTime = newTime

    return(nothing)
    # ---END CODE---
end

function calcMassFlux(  pSurfStart::DArray,         pSurfEnd::DArray,
                        metState::JZMetType,        gridData::JZGridType        )
    # FUNCTION NAME     : calcMassFlux
    # DESCRIPTION       : Calculates air mass fluxes
    # INPUTS:
    #   pSurfStart      : Surface pressure at timestep start (hPa)
    #   pSurfEnd        : Surface pressure at timestep end (hPa)
    #   pGridStart      : Gridded pressures (edge and center) at timestep start (hPa)
    #   pGridEnd        : Gridded pressures (edge and center) at timestep start (hPa)
    #   metState        : Meteorological state
    #   gridData        : Resolution specification and advection factors
    # OUTPUTS:
    #   lonFlux         : E-W flux (hPa)
    #   latFlux         : N-S flux (hPa)
    #   vertFlux        : Vertical flux (hPa)
    #
    # ---START CODE---
    # Establish local references
    gridRel = gridData.gridRel

    # Pressure discrepancy between the start and end of the timestep
    @sync begin
        dPress = [@fetchfrom iWorker begin
            cGridRel = localpart(gridRel)
            cPSS = localpart(pSurfStart)
            cPSE = localpart(pSurfEnd)
            sum((cPSE[:]-cPSS[:]).*cGridRel[:])
        end
       for iWorker in workers()]
    end
    dPress = sum(dPress)

    # dPress is now dgpress
    dpGrid = dfill(0.0,(gridData.nLon,gridData.nLat),workers(),gridData.chunk2D)

    # Adjust pressure grid by mean discrepancy and get the residual
    @sync begin
        [@spawnat iWorker begin
            dPress = remotecall_fetch(myid(),()->dPress)
            cPSS = localpart(pSurfStart)
            cPSE = localpart(pSurfEnd)
            cPSE[:] = cPSE[:] - dPress
            dpGrid = localpart(dpGrid)
            dpGrid[:] = cPSE[:]-cPSS[:]
        end
        for iWorker in workers()]
    end

    # === INIT_PRESS_FIX START ===

    # Re-average pressure at the poles
    avgPole(pSurfStart,gridData)
    avgPole(pSurfEnd,gridData)

    # === SET_PRESS_TERMS START ===

    # Calculate pressure deltas and centers
    (dpStart,dpMid,dpShift) = calcPTerms(pSurfStart,pSurfEnd,gridData)

    # === SET_PRESS_TERMS END ===

    # Copy these over for convenience
    nLon = gridData.nLon
    nLat = gridData.nLat
    nLev = gridData.nLev

    # === CONVERT_WINDS START ===
    # dtdx5 and dtdy5 were already calculated
    # We are using an "A" grid - "C" grid not
    # relevant here and not implemented
    dtdx5 = gridData.dtdx5
    dtdy5 = gridData.dtdy5
    # Courant-Friedrichs-Lewy number in each cell
    UCourant = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    VCourant = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    UWind = metState.UWind
    VWind = metState.VWind

#    # DEBUG
#    @printf("DTD?5: %10.4e:%10.4e\n",dtdx5[4],dtdy5[4])

    @sync begin
        [@spawnat iWorker begin
            cUCourant = localpart(UCourant)
            cVCourant = localpart(VCourant)
            cUWind = localpart(UWind)
            cVWind = localpart(VWind)
            localIdx = myindexes(UWind)
            localLon = localIdx[1]
            localLat = localIdx[2]
            dtdx5 = remotecall_fetch(myid(),()->dtdx5)
            dtdy5 = remotecall_fetch(myid(),()->dtdy5)
            # Ignore the top and bottom latitudes
            nLon = remotecall_fetch(myid(),()->nLon)
            nLat = remotecall_fetch(myid(),()->nLat)
            # First longitude is a special case
            firstLonEl = localLon.==1
            # Establish UWind[iLon] + UWind[iLon+1] first
            for iLat = 1:length(localLat)
                # Both fields ignore the South pole
                targLat = localLat[iLat]
                if targLat != 1
                    # X-field ignores the North pole as well
                    if targLat != nLat
                        cUCourant[2:end,iLat,:] = cUWind[2:end,iLat,:] + cUWind[1:end-1,iLat,:]
                        # Need to retrieve some data from another worker
                        prevLon = localLon[1] - 1
                        if prevLon < 1
                            prevLon = nLon
                        end
                        cUCourant[1,iLat,:] = cUWind[1,iLat,:] + convert(Array,UWind[prevLon,localLat[iLat],:])
                        #for iLev = 1:length(localIdx[3])
                        #    cUCourant[1,iLat,iLev] = cUWind[1,iLat,iLev] + UWind[prevLon,iLat,iLev]
                        #end
                        # Now multiply by the right dtdx5
                        cUCourant[:,iLat,:] = cUCourant[:,iLat,:] .* dtdx5[iLat]
                    end
                    if iLat > 1
                        cVCourant[:,iLat,:] = dtdy5[iLat] .* (cVWind[:,iLat,:] + cVWind[:,iLat-1,:])
                    else
                        # Need data from another worker
                        cVCourant[:,iLat,:] = dtdy5[iLat] .* (cVWind[:,iLat,:] + convert(Array,VWind[localLon,localLat[iLat],:]))
                    end
                end
            end
        end
        for iWorker in workers()]
    end
    
    # === CONVERT_WINDS END ===

    # --- CHECKPOINT: VALIDATED ---

    # === CALC_HORIZ_MASS_FLUX START ===

    lonFlux = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    latFlux = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    cosLatEdge = gridData.cosLatEdge
    nCap = gridData.nCap
    nCap = gridData.nCap
    
#    # DEBUG
#    @printf("Convert UC/VC: %10.4e:%10.4e\n",UCourant[20,20,1],VCourant[20,20,1])

    @sync begin
        [@spawnat iWorker begin
            cLonFlux = localpart(lonFlux)
            cLatFlux = localpart(latFlux)
            cLonWind = localpart(UWind)
            cLatWind = localpart(VWind)
            localIdx = myindexes(latFlux)
            nLon = remotecall_fetch(myid(),()->nLon)
            nCap = remotecall_fetch(myid(),()->nCap)
            localLon = localIdx[1]
            localLat = localIdx[2]
            cdpMid = localpart(dpMid)
            dtdx5 = remotecall_fetch(myid(),()->dtdx5)
            dtdy5 = remotecall_fetch(myid(),()->dtdy5)
            # Fudge this to match what's written in GEOS-Chem
            dtdy5[1] = dtdy5[1+nCap]
            dtdy5[end] = dtdy5[end-nCap]
            dtdx5[1] = dtdx5[1+nCap]
            dtdx5[end] = dtdx5[end-nCap]

            # East-West flux first
            for iLat = 1:size(cLatFlux,2)
                targLat = localLat[iLat]
                xFactor = dtdx5[targLat]
                for iLon = 2:size(cLatFlux,1)
                    cLonFlux[iLon,iLat,:] = (cdpMid[iLon,iLat,:] .* cLonWind[iLon,iLat,:]) + 
                                            (cdpMid[iLon-1,iLat,:].*cLonWind[iLon-1,iLat,:])
                end
                # Forced to communicate with another worker
                iLon = 1
                targLon = localLon[iLon]
                lowLon = targLon - 1
                if lowLon < 1
                    lowLon = nLon
                end
                cLonFlux[iLon,iLat,:] = (cdpMid[iLon,iLat,:] .* cLonWind[iLon,iLat,:]) + 
                    (convert(Array,dpMid[lowLon,localLat[iLat],:]).*convert(Array,UWind[lowLon,localLat[iLat],:]))
                # Convert to an actual flux
                cLonFlux[:,iLat,:] = cLonFlux[:,iLat,:] .* xFactor
            end
            # North-South flux first
            for iLat = 2:length(localLat)
                targLat = localLat[iLat]
                yFactor = dtdy5[targLat] * cosLatEdge[targLat]
                cLatFlux[:,iLat,:] = ((cLatWind[:,iLat,:] .* cdpMid[:,iLat,:]) + 
                                     (cLatWind[:,iLat-1,:].*cdpMid[:,iLat-1,:])).*yFactor
            end
            iLat = 1
            targLat = localLat[iLat]
            yFactor = dtdy5[targLat] * cosLatEdge[targLat]
            if targLat == 1
                cLatFlux[:,iLat,:] = (cLatWind[:,iLat,:] .* cdpMid[:,iLat,:]) .* yFactor
            else
                # Communication
                cLatFlux[:,iLat,:] = ((cLatWind[:,iLat,:] .* cdpMid[:,iLat,:]) + 
                    (convert(Array,VWind[localLon,targLat-1,:]).*
                     convert(Array,dpMid[localLon,targLat-1,:]))).*yFactor
            end
        end
        for iWorker in workers()]
    end

#    # DEBUG
#    @printf("Unfixed lonFlux/latFlux: %10.4e:%10.4e\n",lonFlux[20,20,1],latFlux[20,20,1])

    # --- CHECKPOINT: VALIDATED ---

    # === CALC_HORIZ_MASS_FLUX END ===

    (vertFlux,dpGridFlux) = calcDivergence(lonFlux,latFlux,metState,gridData)

#    # DEBUG
#    @printf("Unfixed vertFlux/dpGridFlux: %10.4e:%10.4e\n",vertFlux[20,20,1],dpGridFlux[20,20])

    # --- CHECKPOINT: VALIDATED ---

    # === INIT_PRESS_FIX END ===

    # === DO_PRESS_FIX_LLNL START ===

    # "Fix" the lonFlux and latFlux
    lonFluxFix = copy(lonFlux)
    latFluxFix = copy(latFlux)

    # Difference between the two discrepancies
    ddpGrid = dfill(0.0,(nLon,nLat),workers(),gridData.chunk2D)
    # Global pressure discrepancy
    dpGlobal = 0.0
    gridRel = gridData.gridRel
    @sync begin
        dpGlobal = 0.0
        [dpGlobal += @fetchfrom iWorker begin
            cddpGrid = localpart(ddpGrid)
            cdpGrid = localpart(dpGrid)
            cdpGridFlux = localpart(dpGridFlux)
            cGridRel = localpart(gridRel)
            cddpGrid[:] = cdpGrid[:] - cdpGridFlux[:]
            cdpGlobal = sum(cddpGrid[:].*cGridRel[:])
            cdpGlobal
        end
        for iWorker in workers()]
    end

#    # DEBUG
#    @printf("ddps/dps/dps_ctm/relA: %10.4e:%10.4e:%10.4e:%10.4e\n",
#            ddpGrid[20,20],dpGrid[20,20],dpGridFlux[20,20],gridRel[20,20])
#
#    gR = convert(Array,gridRel)
#    ddpG = convert(Array,ddpGrid)
#    dpG = convert(Array,dpGrid)
#    dpGF = convert(Array,dpGridFlux)
#    @printf("Alt sum method: %10.4e\n",sum(gR[:].*ddpG[:]))
#    @printf("ddpG[1,2,end-1,end] : %10.4e:%10.4e:%10.4e:%10.4e\n",
#            ddpG[1,1],ddpG[1,2],ddpG[1,end-1],ddpG[1,end])
#    @printf("dpG[1,2,end-1,end]  : %10.4e:%10.4e:%10.4e:%10.4e\n",
#            dpG[1,1],dpG[1,2],dpG[1,end-1],dpG[1,end])
#    @printf("dpGF[1,2,end-1,end] : %10.4e:%10.4e:%10.4e:%10.4e\n",
#            dpGF[1,1],dpGF[1,2],dpGF[1,end-1],dpGF[1,end])
#
#    #@printf("Outputting vertFlux array:\n")
#    #for iLat = 1:nLat
#    #    if mod(iLat,7) == 0
#    #        @printf("\n")
#    #    end
#    #    @printf("%2i %10.4e : ",iLat,maximum(vertFlux[:,iLat,1]))
#    #end
#    #@printf("\n")
#
#    # DEBUG
#    @printf("dpGlobal: %10.4e\n",dpGlobal)

    # Mean meridional flux divergence = dFluxdLat
    @sync begin
        dFluxdLat = zeros(nLat)
        [dFluxdLat += @fetchfrom iWorker begin
            cdFdL = zeros(nLat)
            localIdx = myindexes(ddpGrid)
            cddpGrid = localpart(ddpGrid)
            cdFdL[localIdx[2]] = sum(cddpGrid[:,:],1)
            cdFdL
        end
        for iWorker in workers()]
    end

    dFluxdLat = -((dFluxdLat./nLon) - dpGlobal)

    ## DEBUG
    #@printf("dFdL[20]: %10.4e\n",dFluxdLat[20])

    # Mean meridional flux = meanLonFlux
    meanLonFlux = zeros(nLat)
    meanLonFlux[3] = dFluxdLat[1]/gridData.capFactor
    for iLat in (1+nCap):(nLat-nCap)
        meanLonFlux[iLat+1] = meanLonFlux[iLat] + (dFluxdLat[iLat]/gridData.geoFactor[iLat])
    end

    ## DEBUG
    #@printf("meanLonFlux[20]: %10.4e\n",meanLonFlux[20])

    # This operation will be a nightmare to do in any distributed way
    @sync begin
        fxIntegral = zeros(nLon,nLat)
        [fxIntegral += @fetchfrom iWorker begin
            nLon = remotecall_fetch(myid(),()->nLon)
            nLat = remotecall_fetch(myid(),()->nLat)
            fxMini = zeros(nLon,nLat)
            localIdx = myindexes(ddpGrid)
            cddpGrid = localpart(ddpGrid)
            localLon = localIdx[1]
            localLat = localIdx[2]
            cdpGlobal = remotecall_fetch(myid(),()->dpGlobal)
            cdFdL = remotecall_fetch(myid(),()->dFluxdLat)
            for iLat = 1:length(localLat)
                targLat = localLat[iLat]
                for iLon in 1:length(localLon)
                    targLon = localLon[iLon]
                    fxMini[targLon,targLat] = 
                        (cddpGrid[iLon,iLat] - cdpGlobal) + cdFdL[targLat]
                end
            end
            fxMini
        end
        for iWorker in workers()]
    end

    # Do some local processing for convenience...
    localXCMF = zeros(nLon,nLat)
    for iLat = (1+nCap):(nLat-nCap)
        fxLocalInt = zeros(nLon+1)
        for iLon = 1:nLon
            fxLocalInt[iLon+1] = fxLocalInt[iLon] - fxIntegral[iLon,iLat]
        end
        fxMean = sum(fxLocalInt[2:end])/nLon

        for iLon = 1:nLon
            localXCMF[iLon,iLat] = fxLocalInt[iLon] - fxMean
        end
    end

    ## DEBUG
    #@printf("XCMF[20,20]: %10.4e\n",localXCMF[20,20])

    # Distribute localXCMF -> lonFluxFix
    dpFac = gridData.deltaPFac
    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(lonFluxFix)
            localLon = localIdx[1]
            localLat = localIdx[2]
            clonFluxFix = localpart(lonFluxFix)
            clatFluxFix = localpart(latFluxFix)
            nLev = size(clonFluxFix,3)
            headXCMF = remotecall_fetch(myid(),(lonIdx,latIdx)->localXCMF[lonIdx,latIdx],localLon,localLat)
            dBk = remotecall_fetch(myid(),()->dpFac)
            cmeanLonFlux = remotecall_fetch(myid(),()->meanLonFlux)
            # Distribute vertically
            for iLev = 1:nLev
                # Already contains lonFlux
                clonFluxFix[:,:,iLev] += headXCMF.*dBk[iLev]
                for iLat = 1:length(localLat)
                    targLat = localLat[iLat]
                    clatFluxFix[:,iLat,iLev] += cmeanLonFlux[targLat].*dBk[iLev]
                end
            end
        end
        for iWorker in workers()]
    end
       
    # Re-run divergence calculations
    (vertFluxFix,dpGridFlux) = calcDivergence(lonFluxFix,latFluxFix,metState,gridData)

    # === DO_PRESS_FIX_LLNL END ===

    return(lonFluxFix,latFluxFix,vertFluxFix,dpGridFlux)
    # ---END CODE---
end

function calcDivergence(lonFlux::DArray,latFlux::DArray,metState::JZMetType,gridData::JZGridType)
    # FUNCTION NAME     : calcDivergence
    # DESCRIPTION       : Calculates air mass divergence
    # INPUTS:
    #   lonFlux         : Zonal air flux
    #   latFlux         : Meridional air flux
    #   metState        : Meteorological state variable
    #   gridData        : Resolution specification. Includes
    #                     advection factors
    # OUTPUTS:
    #   vertFlux        : Vertical air flux
    #   dpGridFlux      : New pressure discrepancy (2D)
    #
    # ---START CODE---
    # The divergence is used to calculate vertical motion
    nLon = gridData.nLon
    nLat = gridData.nLat
    nLev = gridData.nLev
    vertFlux = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    geoFactor = gridData.geoFactor
    capFactor = gridData.capFactor
    nCap = gridData.nCap
    nCap = gridData.nCap
    @sync begin
        [@spawnat iWorker begin
            cVert = localpart(vertFlux)
            cLonFlux = localpart(lonFlux)
            cLatFlux = localpart(latFlux)
            # Everything except the polar caps
            localIdx = myindexes(vertFlux)
            localLon = localIdx[1]
            localLat = localIdx[2]
            nLon = remotecall_fetch(myid(),()->nLon)
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            cgeoFac = remotecall_fetch(myid(),()->geoFactor)
            for iLat = 1:length(localLat)
                # Ignore polar caps
                targLat = localLat[iLat]
                if targLat > nCap && targLat < (nLat+1-nCap)
                    # N-S divergence
                    if iLat == length(localLat)
                        # Need to retrieve from another worker
                        cVert[:,iLat,:] = (cLatFlux[:,iLat,:] - convert(Array,latFlux[localLon,targLat+1,:])) .* cgeoFac[iLat]
                    else
                        cVert[:,iLat,:] = (cLatFlux[:,iLat,:] - cLatFlux[:,iLat+1,:]) .* cgeoFac[iLat]
                    end
                    # E-W divergence
                    for iLon = 1:length(localLon)
                        if iLon == length(localLon)
                            targLon = localLon[iLon]+1
                            if targLon > nLon
                                targLon = 1
                            end
                            cVert[iLon,iLat,:] = cVert[iLon,iLat,:] + cLonFlux[iLon,iLat,:] - 
                                       convert(Array,lonFlux[targLon,targLat,:])
                        else
                            cVert[iLon,iLat,:] = cVert[iLon,iLat,:] + cLonFlux[iLon,iLat,:] - 
                                                        cLonFlux[iLon+1,iLat,:]
                        end
                    end
                end
            end
        end
        for iWorker in workers()]
    end

    # === DO_DIVERGENCE_POLE_SUM START ===
    # Set the divergence in the Northernmost and Southernmost latitude bands
    # based simply on influx from/outflux to the next non-polar-cap band
    # Calculate the sum first, average locally
    @sync begin
        SPVec = zeros(2,1,nLev)
        [SPVec += @fetchfrom iWorker begin
            localIdx = myindexes(latFlux)
            localLat = localIdx[2]
            nLat = remotecall_fetch(myid(),()->nLat)
            nLev = length(localIdx[3])
            nCap = remotecall_fetch(myid(),()->nCap)
            returnVec = zeros(2,1,nLev)
            # South pole first
            for iPole in [-1,1]
                if iPole == -1
                    # South Pole
                    targLat = 1 + nCap
                    returnIdx = 1
                elseif iPole == 1
                    targLat = nLat + 1 - nCap
                    returnIdx = 2
                end
                cTargLat = localLat .== targLat
                cLatFlux = localpart(latFlux)
                if any(cTargLat)
                    returnVec[returnIdx,1,:] += sum(cLatFlux[:,findfirst(cTargLat),:],1)
                end
            end
            returnVec
        end
        for iWorker in workers()]
    end

    SPVec = SPVec.*gridData.capFactor./nLon

    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(vertFlux)
            localLat = localIdx[2]
            # Any interest?
            nLat = remotecall_fetch(myid(),()->nLat)
            isSP = any(localLat.==1)
            isNP = any(localLat.==nLat)
            if isSP || isNP
                cVert = localpart(vertFlux)
                localSPV = remotecall_fetch(myid(),()->SPVec)
                localLon = localIdx[1]
                nLev = length(localIdx[3])
                if isSP
                    for iLon = 1:length(localLon)
                        cVert[iLon,1,:] = -localSPV[1,1,:]
                    end
                end
                if isNP
                    for iLon = 1:length(localLon)
                        cVert[iLon,end,:] = localSPV[2,1,:]
                    end
                end
            end

        end
        for iWorker in workers()]
    end


    # === DO_DIVERGENCE_POLE_SUM END ===

    # CTM surface pressure tendency based on flux calculations
    # dpGrid was calcualted based on pSurfs
    dpGridFlux = dfill(0.0,(nLon,nLat),workers(),gridData.chunk2D)

    # Copy 2nd-highest and 2nd-lowest from the polar extremes
    @sync begin
        [@spawnat iWorker begin
            cVert = localpart(vertFlux)
            localIdx = myindexes(vertFlux)
            localLat = localIdx[2]
            nLat = remotecall_fetch(myid(),()->nLat)
            nCap = remotecall_fetch(myid(),()->nCap)
            for iPole in [-1,1]
                if iPole == -1
                    # South pole
                    targLat = nCap
                    sourceLat = 1
                elseif iPole == 1
                    # North pole
                    targLat = nLat + 1 - nCap
                    sourceLat = nLat
                end
                # Do we have the target latitude?
                cTargLat = (localLat .== targLat)
                cSourceLat = (localLat .== sourceLat)
                if any(cTargLat)
                    # If we have the source latitude, no communication req.
                    if any(cSourceLat)
                        cVert[:,findfirst(cTargLat),:] = cVert[:,findfirst(cSourceLat),:]
                    else
                        cVert[:,findfirst(cTargLat),:] = convert(Array,vertFlux[localLon,sourceLat,:])
                    end
                end
            end
        end
        for iWorker in workers()]

    # === CALC_DIVERGENCE END ===

        [@spawnat iWorker begin
            cdpGF = localpart(dpGridFlux)
            cVert = localpart(vertFlux)
            cdpGF[:] = sum(cVert[:,:,:],3)
        end
        for iWorker in workers()]
    end

    return(vertFlux,dpGridFlux)
    # ---END CODE---
end

function calcPTerms(pSurfStart::DArray,pSurfEnd::DArray,gridData::JZGridType)
    # FUNCTION NAME     : calcPTerms
    # DESCRIPTION       : Calculates pressure terms
    # INPUTS:
    #   pSurfStart      : Surface pressure at the start
    #                     of the dynamic timestep
    #   pSurfEnd        : Surface pressure at the end
    #                     of the dynamic timestep
    #   gridData        : Resolution specification
    # OUTPUTS:
    #   dpStart         : Pressure thickness at timestep start (hPa)
    #   dpMid           : Pressure thickness mid-timestep (hPa)
    #   dpShift         : Pressure thickness at mid-longitudes (?)
    #
    # ---START CODE---
    nLon = gridData.nLon
    nLat = gridData.nLat
    nLev = gridData.nLev

    dpStart = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    dpMid   = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    dpShift = dfill(0.0,(nLon,nLat,nLev),workers(),gridData.chunk3D)
    dpOff = gridData.deltaPOff
    dpFac = gridData.deltaPFac

    @sync begin
        [@spawnat iWorker begin
            dAp = remotecall_fetch(myid(),()->dpOff)
            dBk = remotecall_fetch(myid(),()->dpFac)
            nLon = remotecall_fetch(myid(),()->nLon)
            nLev = remotecall_fetch(myid(),()->nLev)

            # Core copies of 3D output grids
            cdpStart = localpart(dpStart)
            cdpMid = localpart(dpMid)
            cdpShift = localpart(dpShift)

            # Core copies of 2D pressure grids
            pSS = localpart(pSurfStart)
            pSE = localpart(pSurfEnd)

            # Temporary - midpoint pressure
            pSM = 0.5.*(pSS[:] + pSE[:])

            # Calculate pressure differences
            for iLev = 1:nLev
                cdpStart[:,:,iLev] = dAp[iLev] + (dBk[iLev].*pSS[:])
                cdpMid[:,:,iLev] = dAp[iLev] + (dBk[iLev].*pSM[:])
            end
        end
        for iWorker in workers()]
    end
   
    @sync begin
        [@spawnat iWorker begin
            cdpMid = localpart(dpMid)
            cdpShift = localpart(dpShift)
            # Careful now..
            localIdx = myindexes(dpMid)
            lonIdx = localIdx[1]
            localLat = localIdx[2]
            localLon = localIdx[1]
            for iLon = 1:length(localLon)
                if iLon > 1
                    cdpShift[iLon,:,:] = 0.5.*(cdpMid[iLon,:,:]+cdpMid[iLon-1,:,:])
                else
                    # Communication required
                    targLon = localLon[iLon]-1
                    if targLon == 0
                        targLon = nLon
                    end
                    cdpShift[iLon,:,:] = 0.5.*(cdpMid[iLon,:,:]+convert(Array,dpMid[targLon,localLat,:]))
                end
            end
#            tempShift = copy(cdpMid)
#            tempShift[2:end,:,:] += tempShift[1:end-1,:,:]
#            targLon = lonIdx[1] - 1
#            if targLon == 0
#                targLon = nLon
#            end
#            # This will need to be fetched from another worker
#            # This process is definitely a bottleneck
#            tempShift[1,:,:] += convert(Array,dpMid[targLon,localLat,:])
#            # Send results to distributed array
#            cdpShift[:] = 0.5.*tempShift[:]
        end
        for iWorker in workers()]
    end

    return(dpStart,dpMid,dpShift)
    # ---END CODE---    
end

# ---END TRANSPORT FUNCTIONS---
