# ---START JULIA ZEPHYR---
# +--------------------------------------------------+
# |     JULIA ZEPHYR (CHEMISTRY) TRANSPORT MODEL     |
# +--------------------------------------------------+
# | The Julia Zephyr (JZ) CTM has been forced into a |
# | single file to deal with some issues which appear|
# | to have resulted from the use of Julia module    |
# | files                                            |
# |                                                  |
# |                                                  |
# +--------------------------------------------------+

# ---START HEADERS---
# New data types
include("JZTypes.jl")

# Transport code (TPCORE from GEOS-Chem)
include("JZTransport.jl")

# ---END HEADERS---
# ---START MAIN FUNCTION---

function runJZephyr(inputFile::ASCIIString)    
    # FUNCTION NAME     : runJZephyr
    # DESCRIPTION       : Main simulation function
    # INPUTS:
    #   inputFile       : Input description
    # OUTPUTS:
    #   runState        : Integer
    #                       0 - Run successful
    #                       1 - Run failed (generic)
    #
    # ---START CODE---

    # Set up constants
    constData = JZConstDataType()

    # Read input options
    (optData,dirData) = readInputFile(inputFile)

    @printf("Running JULIA ZEPHYR, simulation title: %s\n",optData.simName)

    if (stat(dirData.machineFile).size > 0)
        @printf("Setting up processors based on %s...\n",dirData.machineFile)
        # Set up the processors
        setupProc(dirData.machineFile)
    else
        localWorkers = parseint(dirData.machineFile)
        @printf("Setting up %i processors for local run...\n",localWorkers)
        # Set up the processors
        setupProc(dirData.machineFile,localWorkers)
    end

    # Set up time variable
    timeData = JZTimeType(optData)

    @printf("Done. Setting up meteorological state variable...\n") 

    # Set up the relevant data structures
    gridData = JZGridType(optData,constData)
    metState = JZMetType(gridData)

    # Set up tracer data
    tracerData = JZTracGridType(optData,gridData,constData)

    # Initialize output files
    @printf("Setting up output files in %s\n",dirData.outputDir)
    # Overwrite whatever was there before
    MRFiles = Array(IOStream,tracerData.nTracers)
    outputBase = joinpath(abspath(dirData.outputDir),"tracerData")
    # Create directory if necessary
    mkpath(outputBase)

    for iTrc = 1:tracerData.nTracers
        currOut = joinpath(outputBase,tracerData.tracerNames[iTrc])
        MRFiles[iTrc] = open(currOut,"w")
    end

    # Apply initial conditions
    # KLUDGE - Test case
    emitGrid = dfill(0.0,(gridData.nLon,gridData.nLat,tracerData.nTracers),
                            workers(),gridData.chunk3D)
    tempGrid = fill(0.0,size(emitGrid))
    # For debug of PBL mixing
    iLondon = zeros(2)
    for iTracer = 1:tracerData.nTracers
        currName = tracerData.tracerNames[iTracer]
        badTrac = false
        if currName == "NO_LONDON"
            fLon =  [  0.0, 8.0]
            fLat =  [ 51.0,30.0]
        elseif currName == "NO2_HOUSTON"
            fLon = -[ 95.0,23.0]
            fLat =  [ 29.0,46.0]
        elseif currName == "NO3_BEIJING"
            fLon =  [116.0,24.0]
            fLat =  [ 39.0,55.0]
        elseif currName == "N2O5_BOSTON"
            fLon = -[ 71.0, 4.0]
            fLat =  [ 42.0,21.0]
        elseif currName == "O3_SEATTLE"
            fLon = -[122.0,20.0]
            fLat =  [ 47.0,37.0]
        elseif currName == "BC_JOHANNESBURG"
            fLon =  [ 28.0, 3.0]
            fLat = -[ 26.0,12.0]
        else
            badTrac = true
        end
        if !badTrac
            # Convert [deg,min] to decimal degrees
            fLon = (fLon'*[1.0,1.0/60.0])[1]
            # Longitude edges go -182.5 to 177.5
            if fLon > maximum(gridData.lonEdge)
                fLon -= 180.0
            end
            fLat = (fLat'*[1.0,1.0/60.0])[1]
            # Determine the location on the grid
            iLon = findfirst(gridData.lonEdge.>fLon)-1
            iLat = findfirst(gridData.latEdge.>fLat)-1
            # Add a 100 kg/s emission here
            tempGrid[iLon,iLat,iTracer] += 100.0
            if currName == "NO_LONDON"
                iLondon = [iLon iLat]
            end
        end
    end
    #@printf("Emission rate for tracer 1 - local: %10.4f kg/s\n",sum(tempGrid[:,:,1]))
    # Generic routine to assign to distributed grid
    @sync begin
        [@spawnat iWorker begin
            localIdx = myindexes(emitGrid)
            localGrid = localpart(emitGrid)
            localGrid[:] = remotecall_fetch(myid(),()->tempGrid[localIdx...])
        end
        for iWorker in workers()]
    end
    #@printf("Emission rate for tracer 1 - distributed: %10.4f kg/s\n",sum(emitGrid[:,:,1]))

    # Clear the memory taken by tempGrid
    tempGrid = false

    @printf("Determining parameters for initial read operations...\n")
    # Determine the time needed for the first A3 field
    currTimeVec = timeData.tCurr
    A3Time = calcA3Time(currTimeVec,true)
    A6Time = calcA6Time(currTimeVec,true)
    I6Time = calcI6Time(currTimeVec,true)

    @printf("Base time: %04i-%02i-%02i %02i:%02i\n",currTimeVec.year,currTimeVec.month,
                                                    currTimeVec.day,currTimeVec.hour,  
                                                    currTimeVec.minute)
    @printf("A6 time  : %04i-%02i-%02i %02i:%02i\n",A6Time.year,A6Time.month,
                                                    A6Time.day,A6Time.hour,  
                                                    A6Time.minute)
    @printf("I6 time  : %04i-%02i-%02i %02i:%02i\n",I6Time.year,I6Time.month,
                                                    I6Time.day,I6Time.hour,  
                                                    I6Time.minute)
    @printf("A3 time  : %04i-%02i-%02i %02i:%02i\n",A3Time.year,A3Time.month,
                                                    A3Time.day,A3Time.hour,  
                                                    A3Time.minute)

    @printf("Done. Opening first set of met data files...\n") 

    # Open the first set of met data files
    openMetFile(dirData,metState,"I6",I6Time,true)
    openMetFile(dirData,metState,"A6",A6Time,true)
    openMetFile(dirData,metState,"A3",A3Time,true)

    @printf("Done. Populating meteorological state...\n") 

    readA6(metState,A6Time)
    readA3(metState,A3Time)

    # I6 is special - GEOS-Chem has:
    #   GET_I6_FIELDS_1 - Reads into PS1 and TO31
    #   GET_I6_FIELDS_2 - Reads into PS2 and TO32
    # We use an overloaded readI6 function, specifying runStart if necessary
    readI6(metState,I6Time,runStart=true)

    @printf("Initial meteorological state read.\n")

    @printf("Range of PBL depths (m): %6.4f-%6.4f\n",minimum(metState.PBLDepth),maximum(metState.PBLDepth))
    @printf("South Pole pressure limits pre-averaging: %f-%f hPa\n",minimum(metState.pSurfStart[:,1:2]),maximum(metState.pSurfStart[:,1:2]))

    # Average over the poles to deal with the polar singularity problem
    avgPole(metState.pSurfStart,gridData)

    @printf("South Pole pressure limits post-averaging: %f-%f hPa\n",minimum(metState.pSurfStart[:,1:2]),maximum(metState.pSurfStart[:,1:2]))

    # Copy the starting pressure grid to the `current' pressure grid
    metState.pSurf = copy(metState.pSurfStart)

    # Write out the surface pressure and wind arrays for the sake of the report
#    @printf("Writing meteorological data to ASCII files...\n")
#    outputBase = abspath(dirData.outputDir)
#    currOut = joinpath(outputBase,"PSurf")
#    currFile = open(currOut,"w")
#    writeJZArray(metState.pSurf,1,currFile)
#    close(currFile)
#    currOut = joinpath(outputBase,"UWind")
#    currFile = open(currOut,"w")
#    writeJZArray(metState.UWind,1,currFile)
#    close(currFile)
#    currOut = joinpath(outputBase,"VWind")
#    currFile = open(currOut,"w")
#    writeJZArray(metState.VWind,1,currFile)
#    close(currFile)

    # Calculate pressure grids for the first time
    pGridStart = JZPGridType(metState.pSurfStart,gridData)

    # Compute air mass quantities based on the starting surface pressure
    calcAir(metState,pGridStart,gridData,constData)

    # Check that we have something sensible
    @printf("Range of air densities (kg/m3): %6.4f-%6.4f\n",minimum(metState.airDens[:,:,1]),maximum(metState.airDens[:,:,1]))

    # Ignoring PBL effects for now
    # INITIALIZE PBL VARIABLES
    # CALCULATE PBL HEIGHT ETC
    # READ RESTART FILES

    @printf("%s\n","="^80)
    @printf("Initializing master loop\n")
    @printf("%s\n","="^80)
    masterLoopDone = false
    masterFirst = true
    # ===================================================================
    #         *****           M A I N   L O O P             *****
    #         *****         B E G I N S   H E R E           *****
    # ===================================================================
    while !masterLoopDone
    @printf("%s\n","="^80)
    @printf("Beginning 6-hour timestep loop\n")
    @printf("%s\n","="^80)
    # How many minutes left?
    minRemaining = calcTDiff(timeData.tEnd,timeData.tCurr)
    tLoopDyn = min(360,minRemaining)
    nStepDyn = itrunc(tLoopDyn/timeData.tStepDyn)
    updateTime(timeData)
    dynFirst = true
    # ===================================================================
    #         ***** 6 - H O U R   T I M E S T E P   L O O P *****
    #         *****         B E G I N S   H E R E           *****
    # ===================================================================
    for iStepDyn = 1:nStepDyn
        updateTime(timeData)
        tempTime = timeData.tCurr
        @printf("Time step start: %04i-%02i-%02i %02i:%02i:%02i\n",tempTime.year,
            tempTime.month,tempTime.day,tempTime.hour,tempTime.minute,tempTime.second)
        # WRITE DIAGNOSTICS
        # WRITE RESTARTS
#        # Is it time to stop?
#        if timeData.tauCurr >= timeData.tauEnd
#            masterLoopDone = true
#            break
#        end
        # Copy over the previous "end" pressure grids to be the
        # new "start" pressure grids for this timestep
        if ~(dynFirst&&masterFirst)
            pGridStart.pCenter = copy(pGridEnd.pCenter)
            pGridStart.pEdge = copy(pGridEnd.pEdge)
        end
        # Read in new meteorological fields, if necessary
        updateMet(metState,gridData,dirData,timeData,masterFirst&&dynFirst)
        # Interpolate the instantaneous quantities
        interpMet(metState,timeData,gridData)
        # Recalculate the pressure grid based on the timestep end
        pGridEnd = JZPGridType(metState.pSurf,gridData)
        # Recalculate air mass quantities
        calcAir(metState,pGridStart,gridData,constData)
        # Calculate cosine of solar zenith angle
        # calcSZA()

        # Perform emissions based on simple, constant emissions grid
        @printf("Processing emissions...\n")
        tStepSec = float(timeData.tStepDyn)*60.0
        relWeight = tracerData.relWeight
        mixRatio = tracerData.mixRatio
        airMass = metState.airMass
        @sync begin
            [@spawnat iWorker begin
                # Retrieve timestep multiplier
                tStep = remotecall_fetch(myid(),()->tStepSec)
                airLocal = localpart(airMass)
                mixLocal = localpart(mixRatio)
                emitLocal = localpart(emitGrid)
                relWeightLocal = remotecall_fetch(myid(),()->relWeight)
                for iLon in 1:size(airLocal,1)
                    for iLat in 1:size(airLocal,2)
                        kgAir = airLocal[iLon,iLat,1]
                        for iTrac in 1:size(emitLocal,3)
                            # Calculate change in mixing ratio
                            vvEmit = tStep*relWeightLocal[iTrac]*
                                        emitLocal[iLon,iLat,iTrac]/kgAir
                            mixLocal[iLon,iLat,1,iTrac] += vvEmit
                        end
                    end
                end
            end
            for iWorker in workers()]
        end

        # Calculate total tracer mass
        tracSum = calcTracMass(tracerData,metState)
        msgString = "Total tracer mass: "
        for iTrc = 1:tracerData.nTracers
            msgString = @sprintf("%s %10.4e",msgString,tracSum[iTrc])
        end
        @printf("%s kg\n",msgString)

        @printf("Running TPCORE transport module...\n")
        # Run transport
        calcTransport(tracerData,metState,pGridStart,pGridEnd,timeData,gridData,constData)

        # Check vertical mixing
        msgString = "London 1-5:"
        for iLev = 1:5
            msgString = @sprintf("%s %12.4e",msgString,tracerData.mixRatio[iLondon[1],iLondon[2],iLev,1])
        end
        @printf("%s\n",msgString)
        #error("PBLKILL")

        # Calculate total tracer mass
        tracSum = calcTracMass(tracerData,metState)
        msgString = "Total tracer mass: "
        for iTrc = 1:tracerData.nTracers
            msgString = @sprintf("%s %10.4e",msgString,tracSum[iTrc])
        end
        @printf("%s kg\n",msgString)

        # Extremely simplified PBL mixing (GEOS-Chem TURBDAY scheme)
        @printf("Applying PBL mixing...\n")
        dPBLH = metState.PBLDepth # m
        dairMass = metState.airMass
        dpEdge = pGridStart.pEdge
        nTrc = tracerData.nTracers
        dMR = tracerData.mixRatio
        @sync begin
            [@spawnat iWorker begin
                cPBLH = localpart(dPBLH)
                cairMass = localpart(dairMass)
                cpEdge = localpart(dpEdge)
                cMR = localpart(dMR)
                localIdx = myindexes(dairMass)
                localLon = localIdx[1]
                localLat = localIdx[2]
                nLonLocal = length(localLon)
                nLatLocal = length(localLat)
                nLev = length(localIdx[3])
                nTrc = size(cMR,4)
                for iLon in 1:nLonLocal
                    for iLat in 1:nLatLocal
                        # Find max PBL box
                        topPBL = 0
                        pSurf = cpEdge[iLon,iLat,1]
                        pCeil = pSurf
                        # Use barometric law to convert to hPa
                        # Use a scale height of 7,600 m
                        pPBL = pSurf * exp(-cPBLH[iLon,iLat]/7600.0)
                        topFrac = 0.0
                        # Air mass within PBL
                        mPBLAir = 0.0
                        # Tracer mass within PBL
                        mPBLTrc = zeros(nTrc)
                        while pCeil >= pPBL
                            topPBL += 1
                            pCeil = cpEdge[iLon,iLat,topPBL+1]
                            if pCeil > pPBL
                                mPBLAir += cairMass[iLon,iLat,topPBL]
                                for iTrc in 1:nTrc
                                    mPBLTrc[iTrc] += cMR[iLon,iLat,topPBL,iTrc].*cairMass[iLon,iLat,topPBL]
                                end
                            end
                        end
                        if topPBL == 0
                            @printf("%i %f %f!!!2\n",iWorker,pPBL,println(pSurf))
                        end
                        pBase = cpEdge[iLon,iLat,topPBL]
                        # What fraction of the top box is within the PBL?
                        topFrac = 1.0 - ((pPBL - pCeil)/(pBase - pCeil))
                        mPBLAir += (cairMass[iLon,iLat,topPBL]*topFrac)
                        mRatio = zeros(nTrc)
                        for iTrc in 1:nTrc
                            mPBLTrc[iTrc] += (cMR[iLon,iLat,topPBL,iTrc].*cairMass[iLon,iLat,topPBL].*topFrac)
                            mRatio[iTrc] = mPBLTrc[iTrc]./mPBLAir
                        end
                        for iTrc = 1:nTrc
                            for iLev = 1:(topPBL-1)
                                cMR[iLon,iLat,iLev,iTrc] += (mRatio[iTrc]-cMR[iLon,iLat,iLev,iTrc])
                            end
                            cMR[iLon,iLat,topPBL,iTrc] += topFrac.*(mRatio[iTrc]-cMR[iLon,iLat,topPBL,iTrc])
                        end
                        ## DEBUG
                        #if all([localLon[iLon] localLat[iLat]] .== iLondon)
                        #    @printf("topPBL topFrac mPBLAir mPBLTrc mRatio: %10.4f %10.4f %10.4f %10.4f %10.4f\n", topPBL, topFrac, mPBLAir, mPBLTrc[1], mRatio[1])
                        #    miniString = "cMR...: "
                        #    for iLev in 1:topPBL
                        #        miniString = @sprintf("%s%10.4e ",miniString,cMR[iLon,iLat,iLev,1])
                        #    end
                        #    @printf("%s\n",miniString)
                        #end
                    end
                end
            end
            for iWorker in workers()]
        end

        # Check vertical mixing
        msgString = "London 1-5:"
        for iLev = 1:5
            msgString = @sprintf("%s %12.4e",msgString,tracerData.mixRatio[iLondon[1],iLondon[2],iLev,1])
        end
        @printf("%s\n",msgString)
        #error("PBLKILL")

        # Calculate total tracer mass
        tracSum = calcTracMass(tracerData,metState)
        msgString = "Total tracer mass: "
        for iTrc = 1:tracerData.nTracers
            msgString = @sprintf("%s %10.4e",msgString,tracSum[iTrc])
        end
        @printf("%s kg\n",msgString)

        if optData.isDebug || optData.maxOutput
            # Write to the basic diagnostic
            @printf("Writing data to drive (max frequency)...\n")
            for iTrc = 1:tracerData.nTracers
                writeJZArray(tracerData.mixRatio,iTrc,MRFiles[iTrc])
            end
            @printf("Output complete.\n")
        end

        # When running in debug mode, force kill after the first timestep
        if optData.isDebug
            @printf("Debug mode\n")
            @printf("Single step written - killing run\n")
            cleanupJZ(metState,MRFiles)
            return nothing
        end

        # Increment the time elapsed and update the timers
        addTStep(timeData)
        updateTime(timeData)
        dynFirst = false
    end
    # ===================================================================
    #         ***** 6 - H O U R   T I M E S T E P   L O O P *****
    #         *****           E N D S   H E R E             *****
    # ===================================================================
    masterFirst = false
    # Is the loop complete?
    masterLoopDone = (calcTDiff(timeData.tEnd,timeData.tCurr) <= 0)

    # Write to the basic diagnostic
    if !optData.maxOutput
        @printf("Writing data to drive (6 hour frequency)...\n")
        for iTrc = 1:tracerData.nTracers
            writeJZArray(tracerData.mixRatio,iTrc,MRFiles[iTrc])
        end
        @printf("Output complete.\n")
    end

    # DEBUG
    airMass = metState.airMass
    mixRatio = tracerData.mixRatio
    relWeight = tracerData.relWeight
    @sync begin
        sumTrac = [@fetchfrom iWorker begin
            miniSum = 0.0
            airLocal = localpart(airMass)
            mixLocal = localpart(mixRatio)
            relWeightLocal = remotecall_fetch(myid(),()->relWeight)
            for iLon in 1:size(airLocal,1)
                for iLat in 1:size(airLocal,2)
                    kgAir = airLocal[iLon,iLat,1]
                    miniSum += (mixLocal[iLon,iLat,1,1] * kgAir / relWeightLocal[1])
                end
            end
            miniSum
        end
        for iWorker in workers()]
    end
    @printf("Total mass of tracer 1 present: %10.4f\n",sum(sumTrac))

    end
    # ===================================================================
    #         *****           M A I N   L O O P             *****
    #         *****           E N D S   H E R E             *****
    # ===================================================================
    # Main loop complete!

    # Run cleanup
    cleanupJZ(metState,MRFiles)

    @printf("J-Zephyr shutting down.")
    runState = 0
    return(runState,metState,tracerData)
    # ---END CODE---
end

# ---END MAIN FUNCTION---
# ---START INPUT FUNCTIONS---

function readInputFile(inputFile::ASCIIString)
    # FUNCTION NAME     : readInputFile
    # DESCRIPTION       : Reads the input file for J-Zephyr
    # INPUTS:
    #   inputFile       : Input options file
    # OUTPUTS:
    #   optData         : Options structure (JZOptType)
    #   dirData         : Directory data structure (dirData)
    #
    # ---START CODE---
    # Declare function name
    sourceFunc = "readInputFile"

    # OPTIONS INPUT NOT YET READY!
    # Check that input file exists
    if (stat(inputFile).size) == 0
        # Input file does not exist
        errorStop(@sprintf("Input file %s not found",inputFile),sourceFunc)
    end

    # Open the file (read-only)
    inputIOS = open(inputFile,"r")
    lastLine = ""

    # Skip lines until we see the first separator
    while length(lastLine) < 3 || lastLine[1:3] != "---"
        lastLine = readline(inputIOS)
    end

    # We're now at the first separator line
    # Store the rewind point
    lastPos = position(inputIOS)
    while lastLine[1:3] == "---"
        lastPos = position(inputIOS)
        lastLine = readline(inputIOS)
    end

    # Rewind
    seek(inputIOS,lastPos)

    # Read in basic options
    (tempVar,simName    ) = readJZLine(inputIOS,ASCIIString)
    (tempVar,metDir     ) = readJZLine(inputIOS,ASCIIString)
    (tempVar,machineFile) = readJZLine(inputIOS,ASCIIString)
    (tempVar,dStart     ) = readJZLine(inputIOS,Int)
    (tempVar,tStart     ) = readJZLine(inputIOS,Int)
    (tempVar,dEnd       ) = readJZLine(inputIOS,Int)
    (tempVar,tEnd       ) = readJZLine(inputIOS,Int)
    (tempVar,outDir     ) = readJZLine(inputIOS,ASCIIString)
    (tempVar,tStep      ) = readJZLine(inputIOS,Int)
    (tempVar,isDebug    ) = readJZLine(inputIOS,ASCIIString)
    (tempVar,maxOutput  ) = readJZLine(inputIOS,ASCIIString)

    # Parse input
    if isDebug[1] == 'Y'
        isDebug = true
    else
        isDebug = false
    end
    if maxOutput[1] == 'Y'
        maxOutput = true
    else
        maxOutput = false
    end

    # Check that the input file was read correctly
    @printf("Simulation name:    %s\n",simName)
    @printf("Time step:          %d\n",tStep)
    @printf("Met data directory: %s\n",metDir)
    if isDebug
        @printf("Running in debug mode\n")
    end
    if maxOutput
        @printf("Forcing maximum output frequency\n")
    end


    # Is the output directory valid?
    if !ispath(outDir)
        @printf("Output directory does not exist. Attempting to create...\n")
        mkpath(abspath(outDir))
    end

    # Now read in the tracer data
    # Skip lines until we see the first separator
    while length(lastLine) < 3 || lastLine[1:3] != "---"
        lastLine = readline(inputIOS)
    end

    # We're now at the first separator line
    # Store the rewind point
    lastPos = position(inputIOS)
    while lastLine[1:3] == "---"
        lastPos = position(inputIOS)
        lastLine = readline(inputIOS)
    end

    # Read in tracer data
    # First, establish how many tracers we have
    tracerStartPos = lastPos
    nTracers = 0
    while lastLine[1:3] != "---"
        nTracers += 1
        lastLine = readline(inputIOS)
    end
    
    # Rewind
    seek(inputIOS,tracerStartPos)

    # Allocate arrays
    tracerNames = Array(ASCIIString,nTracers)
    molWeight = Array(FloatingPoint,nTracers)

    for iTracer = 1:nTracers
        (tracerNames[iTracer],molWeight[iTracer]) = readJZLine(inputIOS,FloatingPoint)
    end

    # Check the result
    @printf("Details read for %i tracers.\n",nTracers)

    close(inputIOS)

    # Populate the options variables
    optData =JZOptType( simName,    metDir,     outDir,     machineFile,
                        dStart,     dEnd,       tStart,     tEnd,       
                        tStep,      tracerNames,molWeight,  isDebug,     
                        maxOutput                                       )

    # Set up directory data
    dirData = JZDirType(metDir,outDir,machineFile)

    return(optData,dirData)
    # ---END CODE---
end

function readJZLine(inputIOS::IOStream,outType::DataType=ASCIIString)
    # FUNCTION NAME     : readJZLine
    # DESCRIPTION       : Reads a line from the JZ input file
    # INPUTS:
    #   inputIOS        : Input file IO stream
    #   outType         : Optional. Output data type (ASCIIString)
    # OUTPUTS:
    #   outData         : Data read from the line and transformed to the specified type
    #
    # ---START CODE---
    sourceFunc = "readJZLine"
    # Read line from IO stream
    currLine = readline(inputIOS)
    # Cut out everything up to and including first colon, and
    # remove trailing/leading whitespace
    colonLoc = searchindex(currLine,":")
    currLead = strip(currLine[(1:(colonLoc-1))])
    currLine = strip(currLine[(1+colonLoc):end])
    # Convert data if necessary
    if outType != ASCIIString
        if (outType <: Signed) || (outType <: Unsigned)
            outData = parseint(currLine)
        elseif (outType <: FloatingPoint)
            outData = parsefloat(currLine)
        elseif (outType == Bool)
            # Actually reading T or F
            outData = currLine[1] == 'T'
        else
            errorStop("Specified type not recognized",sourceFunc)
        end
        # Convert data type for safety
        outData = convert(outType,outData)
    else
        if outType == Char
            # Convert to a string
            outData = string(outData)
        else
            outData = currLine
        end
    end
    return(currLead,outData)
    # ---END CODE---
end

# ---END INPUT FUNCTIONS---
# ---START OUTPUT FUNCTIONS---
function writeJZArray(outArray::DArray,tracIndex::Int,outFile::IOStream)
    # FUNCTION NAME     : writeJZArray
    # DESCRIPTION       : Very basic routine to output a 3D DArray
    # INPUTS:
    #   outArray        : Output array (2D, 3D or 4D)
    #   tracIndex       : The index of the tracer to be output
    #   outFile         : The file stream to send it to
    # OUTPUTS:
    #   [None]
    #
    # ---START CODE---
    outSize = size(outArray)
    if length(outSize) > 2
        nLev = outSize[3]
    else
        nLev = 1
    end
    for iLev = 1:nLev
        for iLat = 1:outSize[2]
            for iLon = 1:outSize[1]
                if length(outSize) == 2
                    targIdx = [iLon,iLat]
                elseif length(outSize) == 3
                    targIdx = [iLon,iLat,iLev]
                elseif length(outSize) == 4
                    targIdx = [iLon,iLat,iLev,tracIndex]
                end
                @printf(outFile,"%12.6e ",outArray[targIdx...])
            end
            @printf(outFile,"\n")
        end
        @printf(outFile,"\n")
    end
    @printf(outFile,"\n")
    # In case we want to read date mid-simulation
    flush(outFile)
    return(nothing)
    # ---END CODE---
end
# ---END OUTPUT FUNCTIONS---
# ---START MET FUNCTIONS---

function transferMet(metState::JZMetType)
    # FUNCTION NAME     : transferMet
    # DESCRIPTION       : Fields with start/end values in
    #                     the met state are copied from 
    #                     end to start (transitioning to
    #                     new dynamic timestep)
    # INPUTS:
    #   metState        : Meteorological state variable
    # OUTPUTS:
    #   [Modified metState]
    #
    # ---START CODE---
    # Need to transfer:
    #   FROM...     |   TO...
    #---------------+---------------
    # pSurfEnd      | pSurfStart
    # O3DUColEnd    | O3DUColStart
    #---------------+---------------
    pSurfEnd = metState.pSurfEnd
    pSurfStart = metState.pSurfStart
    O3DUColEnd = metState.O3DUColEnd
    O3DUColStart = metState.O3DUColStart
    @sync begin
        [@spawnat iWorker begin
            pSS = localpart(pSurfStart)
            pSE = localpart(pSurfEnd)
            pSS[:] = pSE[:]
            O3S = localpart(O3DUColStart)
            O3E = localpart(O3DUColEnd)
            O3S[:] = O3E[:]
        end
        for iWorker in workers()]
    end
    return(nothing)
    # ---END CODE---
end

function interpMet(metState::JZMetType,timeData::JZTimeType,gridData::JZGridType)
    # FUNCTION NAME     : interpMet
    # DESCRIPTION       : Interpolates instantanoeus (I6) 
    #                     met data
    # INPUTS:
    #   metState        : J-Z meteorological state variable
    #   timeData        : J-Z time variable
    #   gridData        : J-Z grid specification
    # OUTPUTS:
    #   [Modiefied metState]
    #
    # ---START CODE---
    currTime = timeData.tCurr
    # How far are we through the current 6-hour period?
    maxDiff = 60.0*6.0
    # O3 column is interpolated to the timestep mid-point
    # Pressure fields are interpolated to the timestep end
    startDiff = calcTDiff(currTime,calcI6Time(currTime,true))
    midDiff = startDiff + float(timeData.tStepDyn/2.0)
    endDiff = startDiff + float(timeData.tStepDyn)
    midInterpFac = midDiff/maxDiff
    endInterpFac = endDiff/maxDiff
    # Interpolate surface pressure and ozone column
    O3DUColStart = metState.O3DUColStart
    O3DUColEnd = metState.O3DUColEnd
    O3DUCol = metState.O3DUCol
    pSurfStart = metState.pSurfStart
    pSurfEnd = metState.pSurfEnd
    pSurf = metState.pSurf
    pTrop = metState.pTrop
    latEdge = gridData.latEdge
    @sync begin
        [@spawnat iWorker begin
            midInterpFacCore = remotecall_fetch(myid(),()->midInterpFac)
            endInterpFacCore = remotecall_fetch(myid(),()->endInterpFac)
            pSS = localpart(pSurfStart)
            pSE = localpart(pSurfEnd)
            pS = localpart(pSurf)
            O3S = localpart(O3DUColStart)
            O3E = localpart(O3DUColEnd)
            pT = localpart(pTrop)
            O3 = localpart(O3DUCol)
            (nLon,nLat) = size(O3)
            localIdx = myindexes(O3DUCol)
            latCells = localIdx[2]
            latMin = latCells[1]
            latMax = latCells[end]+1
            latEdgeCore = remotecall_fetch(myid(),(lat1,lat2)->latEdge[lat1:lat2],latMin,latMax)
            for iLon = 1:nLon
                for iLat = 1:nLat
                    pS[iLon,iLat] = pSS[iLon,iLat] + (endInterpFacCore*(pSE[iLon,iLat]-pSS[iLon,iLat]))
                    O3[iLon,iLat] = O3S[iLon,iLat] + (midInterpFacCore*(O3E[iLon,iLat]-O3S[iLon,iLat]))
                    if latEdgeCore[iLat] > 60.0 || latEdgeCore[iLat+1] < -60.0
                        pT[iLon,iLat] = max(pT[iLon,iLat],200.0)
                    end
                end
            end
        end
        for iWorker in workers()]
    end
    return(metState)
    # ---END CODE---
end

function updateMet(metState::JZMetType,gridData::JZGridType,dirData::JZDirType,timeData::JZTimeType,firstRun::Bool=false)
    # FUNCTION NAME     : updateMet
    # DESCRIPTION       : Updates the current meteorology,
    #                     if necessary
    # INPUTS:
    #   metState        : J-Z meteorological state variable
    #   gridData        : J-Z grid specification
    #   dirData         : J-Z directory data
    #   timeData        : J-Z time variable
    #   firstRun        : Optional boolean - set to true to force I6 read-in
    # OUTPUTS:
    #   [Modiefied metState]
    #
    # ---START CODE---
    # Test if each field type (A3, I6, A6) needs to be updated in turn
    currTime = timeData.tCurr
    # A3 first
    if ((mod(currTime.hour,3)==0) && (currTime.second==0) && 
                  currTime.minute==0)
        A3Time = calcA3Time(currTime)
        openMetFile(dirData,metState,"A3",A3Time)
        readA3(metState,A3Time)
    end

    # A6 is a little more complex
    A6Date = calcTOffset(currTime,180) 
    if ((mod(A6Date.hour,6)==0) && (A6Date.second==0) && 
                  A6Date.minute==0)
        A6Time = calcA6Time(currTime)
        openMetFile(dirData,metState,"A6",A6Time)
        readA6(metState,A6Time)
    end

    # I6 check is also simple
    if firstRun || ((mod(currTime.hour,6)==0) && (currTime.second==0) && 
                  currTime.minute==0)
        I6Time = calcI6Time(currTime)
        newFile = openMetFile(dirData,metState,"I6",I6Time)
        # Transfer end-of-step fields to start-of-step fields
        # if necessary
        if newFile && ~firstRun
            transferMet(metState)
        end

        readI6(metState,I6Time)

        # Average the pole data
        avgPole(metState.pSurfEnd,gridData)
    end

    return(metState)
    # ---END CODE---
end

function calcAirMass(pGrid::JZPGridType,gridData::JZGridType,constData::JZConstDataType)
    # FUNCTION NAME     : calcAirMass
    # DESCRIPTION       : Dedicated function to calculate an
    #                     air mass grid (kg) based on an
    #                     arbitrary surface pressure dist.
    # INPUTS:
    #   pGrid           : Gridded pressure data
    #   gridData        : Resolution specification
    #   constData       : Constants relevant to J-Zephyr
    # OUTPUTS:
    #   airMass         : Gridded air mass [kg]
    # ---START CODE---
    airMass = dfill(0.0,(gridData.nLon,gridData.nLat,gridData.nLev),workers(),gridData.chunk3D)
    g100 = 100.0/constData.gSurf
    gridArea = gridData.gridArea
    pEdge = pGrid.pEdge
    @sync begin
        [@spawnat iWorker begin
            airLocal = localpart(airMass)
            pEdgeLocal = localpart(pEdge)
            g100 = remotecall_fetch(myid(),()->g100)
            areaLocal = localpart(gridArea)
            for iLev = 1:size(airLocal,3)
                dpBox = pEdgeLocal[:,:,iLev] - pEdgeLocal[:,:,iLev+1]
                airLocal[:,:,iLev] = dpBox .* g100 .* areaLocal
            end
        end
        for iWorker in workers()]
    end
    return airMass
    # ---END CODE---
end

function calcAir(metState::JZMetType,pGrid::JZPGridType,gridData::JZGridType,constData::JZConstDataType)
    # FUNCTION NAME     : calcAir
    # DESCRIPTION       : Calculates box "air" quantities
    # INPUTS:
    #   metState        : Meteorological state variable
    #   pGrid           : Pressure grids
    #   gridData        : Grid data variable
    #   constData       : Constants relevant to J-Z
    # OUTPUTS:
    #   [Modified metState]
    # ---START CODE---
    nLon = gridData.nLon
    # Need what G-C calls Rdg0 = R * 1000 / (g * air (g/mol))
    Rdg0 = constData.RUniv * 1000.0 / (constData.gSurf * constData.airMW)
    g100 = 100.0/constData.gSurf
    # Need to create references to the arrays of interest
    gridAreaOut = gridData.gridArea
    dpBoxOut = metState.dpBox
    dzBoxOut = metState.dzBox
    airVolOut = metState.airVol
    airMassOut = metState.airMass
    airDensOut = metState.airDens
    TMetOut = metState.TMet
    pEdgeOut = pGrid.pEdge
    nLevOut = gridData.nLev
    @sync begin
        [@spawnat iWorker begin
            gridArea = localpart(gridAreaOut)
            dpBox = localpart(dpBoxOut)
            dzBox = localpart(dzBoxOut)
            airVol = localpart(airVolOut)
            airMass = localpart(airMassOut)
            airDens = localpart(airDensOut)
            pEdge = localpart(pEdgeOut)
            TGrid = localpart(TMetOut)
            Rdg0Local = remotecall_fetch(myid(),()->Rdg0)
            g100 = remotecall_fetch(myid(),()->g100)
            nLev = remotecall_fetch(myid(),()->nLevOut)
            localIdx = myindexes(gridAreaOut)
            for iLon=1:length(localIdx[1])
                targLon = localIdx[1][iLon]
                for iLat=1:length(localIdx[2])
                    targLat = localIdx[2][iLat]
                    areaLocal = gridArea[iLon,iLat]
                    for targLev=1:nLev
                        iLev = targLev
                        pBase = pEdge[iLon,iLat,iLev]
                        pCeil = pEdge[iLon,iLat,iLev+1]
                        dpBox[iLon,iLat,iLev] = pBase - pCeil
                        dzBox[iLon,iLat,iLev] = Rdg0Local * TGrid[iLon,iLat,iLev] * log(pBase/pCeil)
                        airVol[iLon,iLat,iLev] = dzBox[iLon,iLat,iLev] * areaLocal
                        airMass[iLon,iLat,iLev] = dpBox[iLon,iLat,iLev] * g100 * areaLocal
                        airDens[iLon,iLat,iLev] = airMass[iLon,iLat,iLev]/airVol[iLon,iLat,iLev]
                    end
                end
            end
        end
        for iWorker in workers()]
    end
    
    return metState
    # ---END CODE---
end

function avgTracerPole(mixRatio::DArray,dpGrid::DArray,gridData::JZGridType)
    # FUNCTION NAME     : avgTracerPole
    # DESCRIPTION       : Averages mixing ratios over the poles
    # INPUTS:
    #   mixRatio        : Full mixing ratio grid
    #   dpGrid          : Pressure thickness grid
    #   gridData        : Grid data structure
    # OUTPUTS:
    #   [Modified mixRatio]
    # ---START CODE---
    # We average over two latitude bands at each pole
    nCap = gridData.nCap
    nLat = gridData.nLat
    for iPole = [-1 1]
        # Perform area-weighted average over both latitude bands
        if iPole == -1
            # South pole
            startPt = 1
        elseif iPole == 1
            startPt = nLat + 1 - nCap
        end
        endPt = startPt + nCap - 1
        capRange = startPt:endPt
        # More generic, (hopefully) less communication
        nLev = size(mixRatio,3)
        nTrc = size(mixRatio,4)
        gridArea = gridData.gridArea
        @sync begin
            # Area sum and value sum
            zonalVals = {zeros(1,1,nLev),zeros(1,1,nLev,nTrc)}
            [zonalVals += @fetchfrom iWorker begin
                localField = localpart(mixRatio)
                localArea = localpart(gridArea)
                localDP = localpart(dpGrid)
                nLev = size(localField,3)
                nTrc = size(localField,4)
                miniVals = {zeros(1,1,nLev),zeros(1,1,nLev,nTrc)}
                localIdx = myindexes(mixRatio)
                localLon = localIdx[1]
                localLat = localIdx[2]
                capRange = remotecall_fetch(myid(),()->capRange)
                for iLat in 1:length(localLat)
                    targLat = localLat[iLat]
                    if any(targLat .== capRange)
                        for iLon in 1:length(localLon)
                            miniVals[1] += localDP[iLon,iLat,:].*localArea[iLon,iLat]
                            for iTrc = 1:nTrc
                                miniVals[2][:,:,:,iTrc] += localField[iLon,iLat,:,iTrc].*localDP[iLon,iLat,:].*localArea[iLon,iLat]
                            end
                        end
                    end
                end
                miniVals
            end
            for iWorker in workers()]
        end
        # Normalize to return the average
        zonalMean = zeros(size(zonalVals[2]))
        for iTrc = 1:nTrc
            zonalMean[1,1,:,iTrc] = zonalVals[2][1,1,:,iTrc]./zonalVals[1][1,1,:]
        end
        @sync begin
            [@spawnat iWorker begin
                # Does this worker have any relevant data?
                localArray = localpart(mixRatio)
                localIdx = myindexes(mixRatio)
                localLon = localIdx[1]
                localLat = localIdx[2]
                capRange = remotecall_fetch(myid(),()->capRange)
                currMean = remotecall_fetch(myid(),()->zonalMean)
                for iLat = 1:length(localLat)
                    targLat = localLat[iLat]
                    isMatch = targLat .== capRange
                    if any(isMatch)
                        for iLon = 1:length(localLon)
                            localArray[iLon,iLat,:,:] = currMean
                        end
                    end
                end
            end
            for iWorker in workers()]
        end
    end

    return mixRatio
    # ---END CODE---
end

function avgPole(avgField::DArray,gridData::JZGridType)
    # FUNCTION NAME     : avgPole
    # DESCRIPTION       : Averages N-D fields over the poles
    # INPUTS:
    #   avgField        : N-D DArray to be averaged over
    #   gridData        : Grid data structure
    # OUTPUTS:
    #   [Modified avgField]
    # ---START CODE---
    # We average over two latitude bands at each pole
    nCap = gridData.nCap
    nLat = gridData.nLat
    for iPole = [-1 1]
        # Perform area-weighted average over both latitude bands
        if iPole == -1
            # South pole
            startPt = 1
        elseif iPole == 1
            startPt = nLat + 1 - nCap
        end
        endPt = startPt + nCap - 1
        capRange = startPt:endPt
        # More generic, (hopefully) less communication
        nLev = size(avgField,3)
        gridArea = gridData.gridArea
        @sync begin
            # Area sum and value sum
            zonalVals = {0.0,zeros(1,1,nLev)}
            [zonalVals += @fetchfrom iWorker begin
                localField = localpart(avgField)
                localArea = localpart(gridArea)
                nLev = size(localField,3)
                miniVals = {0.0,zeros(1,1,nLev)}
                localIdx = myindexes(avgField)
                localLon = localIdx[1]
                localLat = localIdx[2]
                capRange = remotecall_fetch(myid(),()->capRange)
                for iLat in 1:length(localLat)
                    targLat = localLat[iLat]
                    if any(targLat .== capRange)
                        for iLon in 1:length(localLon)
                            miniVals[1] += localArea[iLon,iLat]
                            miniVals[2] += localField[iLon,iLat,:].*localArea[iLon,iLat]
                        end
                    end
                end
                miniVals
            end
            for iWorker in workers()]
        end
        # Normalize to return the average
        zonalMean = zonalVals[2][1,1,:]./zonalVals[1]
        @sync begin
            [@spawnat iWorker begin
                # Does this worker have any relevant data?
                localArray = localpart(avgField)
                localIdx = myindexes(avgField)
                localLon = localIdx[1]
                localLat = localIdx[2]
                capRange = remotecall_fetch(myid(),()->capRange)
                currMean = remotecall_fetch(myid(),()->zonalMean)
                for iLat = 1:length(localLat)
                    targLat = localLat[iLat]
                    isMatch = targLat .== capRange
                    if any(isMatch)
                        for iLon = 1:length(localLon)
                            localArray[iLon,iLat,:] = currMean
                        end
                    end
                end
            end
            for iWorker in workers()]
        end
    end

    return avgField
    # ---END CODE---
end

function openMetFile(dirData::JZDirType,metState::JZMetType,fieldType::String,timeVec::TimeVecType,first::Bool=false)
    # FUNCTION NAME     : openMetFile
    # DESCRIPTION       : Opens an I-6, A-6 or A-3 met file for the given date and
    #                     time. In the example below, reading a hypothetical
    #                     file for 2004-12-01 23:45:30. This particular function
    #                     is overloaded for time vectors
    # INPUTS:
    #   dirData         : Directory and file data
    #   metState        : Meteorological state variable
    #   fieldType       : "A-6", "A-3" or "I-6"
    #   timeVec         : TimeVecType time vector
    #   first           : TRUE if this is the first timestep
    # OUTPUTS:
    #   [Modified metState]
    #   newFile         : Boolean. True if we transitioned to a new file
    #
    # ---START CODE---
    return openMetFile(dirData,metState,fieldType,timeVec.year,timeVec.month,timeVec.day,
                        timeVec.hour,timeVec.minute,timeVec.second,first)
    # ---END CODE---
end

function openMetFile(dirData::JZDirType,metState::JZMetType,fieldType::String,iYear::Int,iMonth::Int,iDay::Int,iHour::Int,iMin::Int,iSec::Int,first::Bool=false)
    # FUNCTION NAME     : openMetFile
    # DESCRIPTION       : Opens an I-6, A-6 or A-3 met file for the given date and
    #                     time. In the example below, reading a hypothetical
    #                     file for 2004-12-01 23:45:30.
    # INPUTS:
    #   dirData         : Directory and file data
    #   metState        : Meteorological state variable
    #   fieldType       : "A-6", "A-3" or "I-6"
    #   iYear           : Integer year (e.g. 2004)
    #   iMonth          : Integer month (e.g. 12 for December)
    #   iDay            : Integer day (e.g. 1)
    #   iHour           : Integer hour (e.g. 23 for 11 pm)
    #   iMin            : Integer minute (e.g. 45)
    #   iSec            : Integer second (e.g. 30)
    #   first           : TRUE if this is the first timestep
    # OUTPUTS:
    #   newFile         : Boolean. True if we transitioned to a new file
    # ---START CODE---
    # For possible errors
    sourceFunc = "openMetFile"
    # Condense the time somewhat
    timeVec = TimeVecType(iYear,iMonth,iDay,iHour,iMin,iSec)
    dateNum = (iYear*10000) + (iMonth*100) + iDay
    timeNum = (iHour*10000) + (iMin*100) + iSec
    # Use dummy function for now
    resString = getResExt()
    # What kind of file are we opening?
    isI6 = (fieldType=="I6")
    isA6 = (fieldType=="A6")
    isA3 = (fieldType=="A3")
    if isI6
        # Needs to be the start of a day
        openFile = (!(compareTime(metState.lastI6FileTime,timeVec))&&(timeNum == 0))
        fileStream = metState.I6File
    elseif isA6
        openFile = (!(compareTime(metState.lastA6FileTime,timeVec))&&(timeNum == 0))
        fileStream = metState.A6File
    elseif isA3
        openFile = (!(compareTime(metState.lastA3FileTime,timeVec)))&&(timeNum == 013000)
        fileStream = metState.A3File
    else
        errorStop(@sprintf("%s not recognized as a met field type",fieldType),sourceFunc)
    end
    newFile = openFile
    # Open at appropriate time or at first simulation tick
    if newFile
        targFile = string("YYYYMMDD.",lowercase(fieldType),".",resString)

        # Need the filepath for the meteorological data
        metPath = dirData.metDir
        # Expand the dates
        targFile = expandDate(joinpath(metPath,targFile),dateNum,timeNum)
        
        # Close any existing files
        # No function ios_isopen (2013-10-26 18:00)
        #if (isa(I6File,IOStream)) && isopen(I6File)
        close(fileStream)

        # Does the file exist?
        if !isfile(targFile)
            errorStop(@sprintf("%s does not exist",targFile),sourceFunc);
        end

        # Can we read it?
        if !isreadable(targFile)
            errorStop(@sprintf("%s could not be read",targFile),sourceFunc);
        end

        # Open the file
        fileStream = open(targFile)

        # Report that the file is open
        @printf("%s file stream linked to %s\n",fieldType,targFile)

        # Read the identification string
        identString = bytestring(readFORTRANRecord(fileStream,Uint8))

        # Could error check here - string should read [G5 45 NN]
        # Corresponds to GEOS-5 fields, 4x5 degree, NN stored

        # Last 2 characters are the number of fields
        nFields = parseint(identString[7:8]);

        if isI6
            metState.I6File = fileStream
            metState.nI6Fields = nFields
            timeVec.hour = 0
            timeVec.minute = 0
            timeVec.second = 0
            metState.lastI6FileTime = copyTime(timeVec)
        elseif isA6
            metState.A6File = fileStream
            # An error in the source data means we have to kludge this
            # 2 fields are present in the file, but not counted in the header
            if (nFields != 19)
                errorStop("Wrong number of fields in the A-6 file",sourceFunc)
            end
            nFields = 21
            metState.nA6Fields = nFields
            timeVec.hour = 0
            timeVec.minute = 0
            timeVec.second = 0
            metState.lastA6FileTime = copyTime(timeVec)
        elseif isA3
            metState.A3File = fileStream
            metState.nA3Fields = nFields
            timeVec.hour = 1
            timeVec.minute = 30
            timeVec.second = 0
            metState.lastA3FileTime = copyTime(timeVec)
        end
    end
    return(newFile)
    # ---END CODE---
end

function getResExt()
    # FUNCTION NAME     : getResExt
    # DESCRIPTION       : Dummy function. Returns the current
    #                     grid resolution as a string, such as
    #                     "4x5" if using a 4x5 grid.
    # INPUTS:
    # OUTPUTS:
    #   resString       : Resolution string
    #
    # Output variable declarations
    # Local variable declarations
    # ---START CODE---
    resString = "4x5"
    return(resString)
    # ---END CODE---
end

function readA3(metState::JZMetType,timeVec::TimeVecType)
    return readA3(metState,timeVec.year,timeVec.month,timeVec.day,timeVec.hour,timeVec.minute,timeVec.second)
end

function readA3(metState::JZMetType,iYear::Int,iMonth::Int,iDay::Int,iHour::Int,iMin::Int,iSec::Int)
    # FUNCTION NAME     : readA3
    # DESCRIPTION       : Reads all A3 file data into the 
    #                     relevant variableses.
    # INPUTS:
    #   metState        : Meteorological state variable
    #   iYear           : Integer year
    #   iMonth          : Integer month
    #   iDay            : Integer day
    #   iHour           : Integer hour
    #   iMin            : Integer minute
    #   iSec            : Integer second
    # OUTPUTS:
    # ---START CODE---
    sourceFunc = "readA3"
    # Count number of fields retrieved
    nFound = 0
    # Calculate the current datenum and timenum for comparison
    timeVec = TimeVecType(iYear,iMonth,iDay,iHour,iMin,iSec)
    currYMD = DTIntJZ(iYear,iMonth,iDay)
    currHMS = DTIntJZ(iHour,iMin,iSec)
    stopRead = compareTime(timeVec,metState.lastA3ReadTime)
    if stopRead
        @printf("A3 file already read in for %04i-%02i-%02i %02i:%02i\n",
            timeVec.year,timeVec.month,timeVec.day,timeVec.hour,timeVec.minute)
        return(true)
    else
        @printf("Reading A3 meteorology for %04i-%02i-%02i %02i:%02i\n",
            timeVec.year,timeVec.month,timeVec.day,timeVec.hour,timeVec.minute)
    end
    while !stopRead
        # Check for EOF
        stopRead = eof(metState.A3File)
        if !stopRead
            # Get the current field name
            fieldName = uppercase(rstrip(bytestring(readFORTRANRecord(metState.A3File,Uint8))))
            # Hack with IF-ELSE for now
            nameOK = true
            addExtra = false
            if (fieldName == "ALBEDO")
                targArray = metState.sfcAlbedo
            elseif (fieldName == "CLDTOT") || (fieldName == "CLDFRC")
                targArray = metState.cldFrc2D
            elseif (fieldName == "EFLUX")
                targArray = metState.EFlux
            elseif (fieldName == "FRLAKE")
                targArray = metState.fracLake
            elseif (fieldName == "FRLAND")
                targArray = metState.fracLand
            elseif (fieldName == "FROCEAN")
                targArray = metState.fracOcean
            elseif (fieldName == "FRLANDIC")
                targArray = metState.fracLandIce
            elseif (fieldName == "EVAP")
                targArray = metState.evapFlux
            elseif (fieldName == "GRN")
                targArray = metState.greenIdx
            elseif (fieldName == "GWETROOT")
                targArray = metState.wetRoot
            elseif (fieldName == "GWETTOP") || (fieldName == "GWETSOIL")
                targArray = metState.wetTop
            elseif (fieldName == "HFLUX")
                targArray = metState.sensFlux
            elseif (fieldName == "LAI")
                targArray = metState.LAI
            elseif (fieldName == "PARDF") || (fieldName == "PARDIF")
                targArray = metState.PARDiffuse
            elseif (fieldName == "PARDR") || (fieldName == "PARDIR")
                targArray = metState.PARDirect
            elseif (fieldName == "PBLH") || (fieldName == "PBLH")
                targArray = metState.PBLDepth
            elseif (fieldName == "PRECANV")
                targArray = metState.prcpAnvSurf
            elseif (fieldName == "PRECTOT") || (fieldName == "PREACC")
                targArray = metState.prcpTotSurf
            elseif (fieldName == "PRECCON") || (fieldName == "PRECON")
                targArray = metState.prcpConSurf
            elseif (fieldName == "PRECSNO")
                targArray = metState.prcpSnoSurf
            elseif (fieldName == "LWGNET") || (fieldName == "RADLWG")
                targArray = metState.netRadLW
            elseif (fieldName == "LWTUP")
                targArray = metState.upRadLW
            elseif (fieldName == "QV2M")
                targArray = metState.specHum2m
            elseif (fieldName == "SWGNET") || (fieldName == "RADSWG")
                targArray = metState.netRadSW
            elseif (fieldName == "SNODP")
                targArray = metState.snowGeo
            elseif (fieldName == "SNOMAS")
                targArray = metState.snowH2O
            elseif (fieldName == "TROPP")
                targArray = metState.pTrop
            elseif (fieldName == "T2M") || (fieldName == "TS") || (fieldName == "TGROUND")
                targArray = metState.T2m
            elseif (fieldName == "TSKIN")
                targArray = metState.TSkin
            elseif (fieldName == "U10M") || (fieldName == "USS")
                targArray = metState.U10m
            elseif (fieldName == "USTAR")
                targArray = metState.UStar
            elseif (fieldName == "V10M") || (fieldName == "VSS")
                targArray = metState.V10m
            elseif (fieldName == "Z0") || (fieldName == "Z0M")
                targArray = metState.Z0
            else
                # Field not relevant
                nameOK = false
            end
            # Read info
            (tempData,tempDate,tempTime) = readFORTRANRecord(metState.A3File,Float32,true)
            if nameOK
                # Does the date match?
                dateOK = (currYMD == itrunc(tempDate)) && (currHMS == itrunc(tempTime))
                if dateOK
                    # Reshape the data
                    tempData = reshape(tempData,size(targArray))
                    # Read data into target field
                    @sync begin
                        [@spawnat iWorker begin
                            targLocal = localpart(targArray)
                            targIdx = myindexes(targArray)
                            targLocal[:] = remotecall_fetch(myid(),()->tempData[targIdx...])
                        end
                        for iWorker in workers()]
                    end
                    nFound += 1
                end
            end
            stopRead = stopRead | (nFound == metState.nA3Fields)
        else
            # Run final checks
            if (nFound != metState.nA3Fields)
                # Error - reached end of file but did not find all fields!
                close(metState.A3File)
                errorStop(@sprintf("Missing %d of %d A-3 fields for %04d-%02d-%02d %02d:%02d",
                    (metState.nA3Fields-nFound),metState.nA3Fields,iYear,iMonth,iDay,iHour,iMin),sourceFunc)
            end
        end
    end
    # Update last read time
    metState.lastA3Time = currHMS
    metState.lastA3Date = currYMD
    metState.lastA3ReadTime = copyTime(timeVec)
    # Convert GEOS-5 precipitation fields to older convention, and correct GEOS-5
    # PBL depth to ensure that it is >= minimum mechanical mixing depth
    prcpAnvSurf = metState.prcpAnvSurf
    prcpTotSurf = metState.prcpTotSurf
    prcpConSurf = metState.prcpConSurf
    PBLDepth = metState.PBLDepth;
    UStar = metState.UStar;
    @sync begin
        [@spawnat iWorker begin
            prcpLocal = localpart(prcpAnvSurf)
            prcpLocal[:] *= 86400.0
            prcpLocal = localpart(prcpTotSurf)
            prcpLocal[:] *= 86400.0
            prcpLocal = localpart(prcpConSurf)
            prcpLocal[:] *= 86400.0
            PBLLocal = localpart(PBLDepth)
            UStarLocal = localpart(UStar)
            for iEl = 1:length(PBLLocal)
                minDepth = 700.0*UStarLocal[iEl]
                PBLLocal[iEl] = max(PBLLocal[iEl],minDepth)
            end
        end
        for iWorker in workers()]
    end

    return(nFound==metState.nA3Fields)
    # ---END CODE---
end

function readA6(metState::JZMetType,timeVec::TimeVecType)
    return readA6(metState,timeVec.year,timeVec.month,timeVec.day,timeVec.hour,timeVec.minute,timeVec.second)
end

function readA6(metState::JZMetType,iYear::Int,iMonth::Int,iDay::Int,iHour::Int,iMin::Int,iSec::Int)
    # FUNCTION NAME     : readA6
    # DESCRIPTION       : Reads all A6 file data into the 
    #                     relevant variableses.
    # INPUTS:
    #   metState        : Meteorological state
    #   iYear           : Integer year
    #   iMonth          : Integer month
    #   iDay            : Integer day
    #   iHour           : Integer hour
    #   iMin            : Integer minute
    #   iSec            : Integer second
    # OUTPUTS:
    # ---START CODE---
    # Grid sizes
    (nLon,nLat,nLev) = getGridSize()
    sourceFunc = "readA6"
    # Count number of fields retrieved
    nFound = 0
    # Calculate the current datenum and timenum for comparison
    timeVec = TimeVecType(iYear,iMonth,iDay,iHour,iMin,iSec)
    currYMD = DTIntJZ(iYear,iMonth,iDay)
    currHMS = DTIntJZ(iHour,iMin,iSec)
    stopRead = compareTime(timeVec,metState.lastA6ReadTime)
    if stopRead
        @printf("A6 file already read in for %04i-%02i-%02i %02i:%02i\n",
            timeVec.year,timeVec.month,timeVec.day,timeVec.hour,timeVec.minute)
        return(true)
    else
        @printf("Reading A6 meteorology for %04i-%02i-%02i %02i:%02i\n",
            timeVec.year,timeVec.month,timeVec.day,timeVec.hour,timeVec.minute)
    end
    while !stopRead
        # Check for EOF
        stopRead = eof(metState.A6File)
        if !stopRead
            # Get the current field name
            fieldName = uppercase(rstrip(bytestring(readFORTRANRecord(metState.A6File,Uint8))))
            # Equivalent to SELECT CASE ( NAME )?

            # Hack with IF-ELSE for now
            nameOK = true
            addExtra = false
            if (fieldName == "CLOUD")
                targArray = metState.cldFrc3D
            elseif (fieldName == "DTRAIN")
                targArray = metState.detrainFlux
            elseif (fieldName == "CMFMC")
                targArray = metState.convectFlux
                # Finding convectFlux allows us to derive cldTop
                addExtra = true
            elseif (fieldName == "DQIDTMST")
                targArray = metState.iceTendMoist
            elseif (fieldName == "DQLDTMST")
                targArray = metState.liqTendMoist
            elseif (fieldName == "DQVDTMST")
                targArray = metState.vapTendMoist
            elseif (fieldName == "DQRCON")
                targArray = metState.prcpConForm
            elseif (fieldName == "DQRLSC")
                targArray = metState.prcpLrgForm
            elseif (fieldName == "OPTDEPTH")
                targArray = metState.tauCld
            elseif (fieldName == "TAUCLI")
                targArray = metState.tauIce
            elseif (fieldName == "TAUCLW")
                targArray = metState.tauLiq
            elseif (fieldName == "RH")
                targArray = metState.relHum
            elseif (fieldName == "QV")
                targArray = metState.specHum
            elseif (fieldName == "QL")
                targArray = metState.liqRatio
            elseif (fieldName == "QI")
                targArray = metState.iceRatio
            elseif (fieldName == "U")
                targArray = metState.UWind
            elseif (fieldName == "V")
                targArray = metState.VWind
            elseif (fieldName == "OMEGA")
                targArray = metState.pVelocity
            elseif (fieldName == "T")
                targArray = metState.TMet
            elseif (fieldName == "MOISTQ")
                targArray = metState.vapTend
            else
                # Field not relevant
                nameOK = false
            end
            # Read info
            (tempData,tempDate,tempTime) = readFORTRANRecord(metState.A6File,Float32,true)
            if nameOK
                # Does the date match?
                dateOK = (currYMD == itrunc(tempDate)) && (currHMS == itrunc(tempTime))
                if dateOK
                    # Reshape the data
                    tempData = reshape(tempData,size(targArray))
                    # Read data into target field
                    @sync begin
                        [@spawnat iWorker begin
                            targLocal = localpart(targArray)
                            targIdx = myindexes(targArray)
                            targLocal[:] = remotecall_fetch(myid(),()->tempData[targIdx...])
                        end
                        for iWorker in workers()]
                    end
                    nFound += 1
                    if addExtra
                        nFound += 1
                    end
                end
            end
            stopRead = stopRead | (nFound == metState.nA6Fields)
        else
            if (nFound != metState.nA6Fields)
                # Error - reached end of file but did not find all fields!
                close(metState.A6File)
                errorStop(@sprintf("Missing %d of %d A-6 fields for %04d-%02d-%02d %02d:%02d",
                    (metState.nA6Fields-nFound),metState.nA6Fields,iYear,iMonth,iDay,iHour,iMin),sourceFunc)
            end
        end
    end

    # Update last read time
    metState.lastA6Time = currHMS
    metState.lastA6Date = currYMD
    metState.lastA6ReadTime = copyTime(timeVec)
    # Test update - this definitely works
    tempVap  = metState.vapTend
    tempTop  = metState.cldTop
    tempRH   = metState.relHum
    tempSH   = metState.specHum
    tempFlux = metState.convectFlux
    @sync begin
        [@spawnat iWorker begin
            # Calculate cloud top level
            localTop = localpart(tempTop)
            localFlux = localpart(tempFlux)
            (nLon,nLat,nLev) = size(localFlux)
             for iLon in 1:nLon
                 for iLat in 1:nLat
                     iLev = nLev
                     while !(localFlux[iLon,iLat,iLev] > 0.0 || iLev == 1)
                         iLev -= 1
                     end
                     localTop[iLon,iLat] = iLev + 1
                 end
             end
            # Convert vapor tendency
            targLocal = localpart(tempVap)
            targLocal[:] = targLocal / -8.64e7
            # Convert relative humidity from fraction to %
            targLocal = localpart(tempRH)
            targLocal[:] = targLocal * 100.0
            # Convert specific humidity from kg/kg to g/kg
            targLocal = localpart(tempSH)
            targLocal[:] = targLocal * 1000.0
        end
        for iWorker in workers()]
    end

    return(nFound==metState.nA6Fields)
    # ---END CODE---
end

function readI6(metState::JZMetType,timeVec::TimeVecType;runStart::Bool=false)
    return readI6(metState,timeVec.year,timeVec.month,timeVec.day,timeVec.hour,timeVec.minute,timeVec.second,runStart=runStart)
end

function readI6(metState::JZMetType,iYear::Int,iMonth::Int,iDay::Int,iHour::Int,iMin::Int,iSec::Int;runStart::Bool=false)
    # FUNCTION NAME     : readI6
    # DESCRIPTION       : Reads all I6 file data into the 
    #                     relevant variableses.
    # INPUTS:
    #   metState        : Meteorological state
    #   iYear           : Integer year
    #   iMonth          : Integer month
    #   iDay            : Integer day
    #   iHour           : Integer hour
    #   iMin            : Integer minute
    #   iSec            : Integer second
    # OUTPUTS:
    # ---START CODE---
    sourceFunc = "readI6"
    # Count number of fields retrieved
    nFound = 0
    # Calculate the current datenum and timenum for comparison
    timeVec = TimeVecType(iYear,iMonth,iDay,iHour,iMin,iSec)
    currYMD = DTIntJZ(iYear,iMonth,iDay)
    currHMS = DTIntJZ(iHour,iMin,iSec)
    stopRead = compareTime(timeVec,metState.lastI6ReadTime)
    if stopRead
        @printf("I6 file already read in for %04i-%02i-%02i %02i:%02i\n",
            timeVec.year,timeVec.month,timeVec.day,timeVec.hour,timeVec.minute)
        return(true)
    else
        @printf("Reading I6 meteorology for %04i-%02i-%02i %02i:%02i\n",
            timeVec.year,timeVec.month,timeVec.day,timeVec.hour,timeVec.minute)
    end
    while !stopRead
        # Check for EOF
        stopRead = eof(metState.I6File)
        if !stopRead
            # Get the current field name
            fieldName = uppercase(rstrip(bytestring(readFORTRANRecord(metState.I6File,Uint8))))
            # Equivalent to SELECT CASE ( NAME )?
            nameOK = true
            if (fieldName == "LWI")
                targArray = metState.LWI
            elseif (fieldName == "PS")
                if runStart
                    targArray = metState.pSurfStart
                else
                    targArray = metState.pSurfEnd
                end
            elseif (fieldName == "SLP")
                targArray = metState.pSeaLevel
            elseif (fieldName == "TO3")
                if runStart
                    targArray = metState.O3DUColStart
                else
                    targArray = metState.O3DUColEnd
                end
            elseif (fieldName == "TTO3")
                targArray = metState.O3DUTrop
            else
                # Field not relevant
                nameOK = false
            end
            # Read info
            (tempData,tempDate,tempTime) = readFORTRANRecord(metState.I6File,Float32,true)
            if nameOK
                # Does the date match?
                dateOK = (currYMD == itrunc(tempDate)) && (currHMS == itrunc(tempTime))
                if dateOK
                    # Reshape the data
                    tempData = reshape(tempData,size(targArray))
                    # Read data into target field
                    @sync begin
                        [@spawnat iWorker begin
                            targLocal = localpart(targArray)
                            targIdx = myindexes(targArray)
                            targLocal[:] = remotecall_fetch(myid(),()->tempData[targIdx...])
                        end
                        for iWorker in workers()]
                    end
                    nFound += 1
                end
            end
            stopRead = stopRead | (nFound == metState.nI6Fields)
        else
            if (nFound != metState.nI6Fields)
                # Error - reached end of file but did not find all fields!
                close(metState.I6File)
                errorStop(@sprintf("Missing %d of %d I-6 fields for %04d-%02d-%02d %02d:%02d",
                    (metState.nI6Fields-nFound),metState.nI6Fields,iYear,iMonth,iDay,iHour,iMin),sourceFunc)
            end
        end
    end
    # Update last read time
    metState.lastI6Time = currHMS
    metState.lastI6Date = currYMD
    metState.lastI6ReadTime = copyTime(timeVec)

    return(nFound==metState.nI6Fields)
    # ---END CODE---
end

function closeAll(metState::JZMetType)
    # FUNCTION NAME     : closeAll
    # DESCRIPTION       : Closes all open met data file
    #                     streams (A3, I6 and A6)
    # INPUTS:
    #   metState        : Meteorological state vector
    # OUTPUTS:
    #
    # ---START CODE---
    close(metState.I6File)
    close(metState.A6File)
    close(metState.A3File)

    return(nothing)
    # ---END CODE---
end

# ---END MET FUNCTIONS---
# ---START IO FUNCTIONS---

function htonArr(xNet::Array)
    # FUNCTION NAME     : htonArr                  
    # DESCRIPTION       : Converts all values in an
    #                     array to be big-endian
    # INPUTS:
    #   xNet            : Array (host-endian)
    # OUTPUTS:
    #   xBEnd           : Array converted to be big-endian
    #
    # ---START CODE---
    xBEnd = copy(xNet)
    htonArr!(xBEnd)
    return(xBEnd)
    # ---END CODE---
end

function htonArr!(xNet::Array)
    # FUNCTION NAME     : htonArr!
    # DESCRIPTION       : Converts all values in an
    #                     array to be big-endian (in-place)
    # INPUTS:
    #   xNet            : Array (host-endian)
    # OUTPUTS:
    #   xNet            : Array converted to be big-endian
    #
    # ---START CODE---
    for iVal=1:length(xNet)
        xNet[iVal] = hton(xNet[iVal])
    end
    return(xNet)
    # ---END CODE---
end

function readFORTRANRecord(IOS::IOStream,typeOut::Type=Uint8,getDateHeader::Bool=false)
    # FUNCTION NAME     : readFORTRANRecord
    # DESCRIPTION       : Reads a FORTRAN record
    # INPUTS:
    #   IOS             : IO stream to read from
    #   typeOut         : Type of data to read
    #   getDateHeader   : Read 2 Uint32 values at the start as the date and time
    # OUTPUTS:
    #   xNet            : Array converted to be big-endian
    #
    # ---START CODE---
    # What is our reducing factor?
    inPerOut = sizeof(typeOut)
    # Read the first 4 bytes to determine record length
    recLen = itrunc(hton(read(IOS,Uint32)))
    # Read data
    if getDateHeader
        # Read 2 Uint32 values at the start
        recLenAdj = recLen - 8
        dateNum = hton(read(IOS,Uint32))
        timeNum = hton(read(IOS,Uint32))
    else
        dateNum = convert(Uint32,0)
        timeNum = convert(Uint32,0)
        recLenAdj = recLen
    end
    nRecords = itrunc(recLenAdj/inPerOut)
    data = htonArr(read(IOS,typeOut,nRecords))
    # Check that we actually got what we expected
    recLenCheck = hton(read(IOS,Uint32))
    if recLenCheck != recLen
        data = []
        @printf("Insert read failure")
    end
    if getDateHeader
        return(data,dateNum,timeNum)
    else
        return(data)
    end
    # ---END CODE---
end

# ---END IO FUNCTIONS---
# ---START PROC FUNCTIONS---
function setupProc(machineFile,localWorkers::Int=5)
    # FUNCTION NAME     : setupProc
    # DESCRIPTION       : Initializes the worker processes
    #                     and calculates how to distribute
    #                     the various arrays
    # INPUTS:
    #   localWorkers    : Number of workers to use if
    #                     running locally
    # OUTPUTS:
    #
    # ---START CODE---
    # Target variables
    # First, read in the machine file
    (nodeWorkers,nodeIP,nZHost) = readMachineList(machineFile,localWorkers)
    ZHostIP=nodeIP
    if nZHost==1
        # Local
        ZWorkerID=workers()
    else
        # Distributed
        ZWorkerID=nodeWorkers
    end
    return(ZWorkerID)
    # ---END CODE---
end

function readMachineList(machineListFile::String,localWorkers::Int=5)
    # FUNCTION NAME     : readMachineList
    # DESCRIPTION       : Initializes the worker processes
    #                     and calculates how to distribute
    #                     the various arrays
    # INPUTS:
    #   machineListFile : String describing the absolute
    #                     path of the machine list file
    #   localWorkers    : Number of workers to use if
    #                     running locally
    # OUTPUTS:
    #   nodeWorkers     : Array containing the number of
    #                     workers to be established on
    #                     each node
    #   nodeIP          : IP address of each node
    #
    # ---START CODE---
    if !isfile(machineListFile)
        # Run in local parallel mode
        currWorkers = nworkers()
        maxWorkers = localWorkers
        if currWorkers < maxWorkers
            if currWorkers == nprocs() == 1
                currWorkers = 0
            end
            addprocs(maxWorkers-currWorkers)
        end
        nZWorker=nworkers()
        nZHost=1
        nodeWorkers=nworkers()
        nodeIP=fill(ip"127.0.0.1",nodeWorkers)
    else
        mlistStream = open(machineListFile,read)
        # Set the number of workers
        error("Machine list reading routine not ready")
    end

    return(nodeWorkers,nodeIP,nZHost)
    # ---END CODE---
end

# ---END PROCESSOR FUNCTIONS---
# ---START PARAMETER FUNCTIONS---

function getGridSize()
    # FUNCTION NAME     : getGridSize
    # DESCRIPTION       : Returns the overall grid size
    # INPUTS:
    # OUTPUTS:
    #   nLon            : Number of longitudinal cells
    #   nLat            : Number of latitudinal cells
    #   nLev            : Number of levels
    #
    # ---START CODE---
    nLon = 72
    nLat = 46
    nLev = 72
    return(nLon,nLat,nLev)
    # ---END CODE---
end

function getResExt()
    # FUNCTION NAME     : getResExt
    # DESCRIPTION       : Dummy function. Returns the current
    #                     grid resolution as a string, such as
    #                     "4x5" if using a 4x5 grid.
    # INPUTS:
    # OUTPUTS:
    #   resString       : Resolution string
    #
    # Output variable declarations
    # Local variable declarations
    # ---START CODE---
    resString = "4x5"
    return(resString)
    # ---END CODE---
end

# ---END PARAMETER FUNCTIONS---
# ---START ERROR FUNCTIONS---

function errorStop(stopMsg::String="Unknown error",sourceFunc::String="unknown function")
    # FUNCTION NAME     : JZStop
    # DESCRIPTION       : Safely stops J-Zephyr after an error
    #                     occurs in some source subroutine
    # INPUTS:
    #   stopMsg         : Message to print to STDOUT
    #   sourceFunc      : Optional - source routine
    # OUTPUTS:
    # ---START CODE---
    # Let the user know that something is wrong
    @printf("%s\n","="^80)
    @printf("ERROR: J-Z stopped in %s\n%s\n",sourceFunc,stopMsg)
    @printf("%s\n","="^80)
    # Call the standard stop routine
    #mainMod.JZStop()
    error()
    return(nothing)
    # ---END CODE---
end

# ---END ERROR FUNCTIONS---
# ---START TIME FUNCTIONS---
function compareTime(timeA::TimeVecType,timeB::TimeVecType)
    return (timeA.year==timeB.year&&timeA.month==timeB.month&&timeA.day==timeB.day&&timeA.hour==timeB.hour&&timeA.minute==timeB.minute&&timeA.second==timeB.second)
end

function copyTime(timeVec::TimeVecType)
    return TimeVecType(timeVec.year,timeVec.month,timeVec.day,timeVec.hour,timeVec.minute,timeVec.second)
end

function updateTime(timeData::JZTimeType)
    # FUNCTION NAME     : updateTime
    # DESCRIPTION       : Update the current time in line
    #                     with the number of minutes elapsed
    # INPUTS:
    #   timeData        : JZTimeType time data
    # OUTPUTS:
    #   [Modified time data]
    #
    # ---START CODE---
    timeData.tauCurr = calcTau(timeData.tCurr)
    return(timeData)
    # ---END CODE---
end

function calcFracDay(timeVec::TimeVecType)
    # FUNCTION NAME     : calcFracDay
    # DESCRIPTION       : Calculates the fractional number of days
    #                     which have passed in the current month.
    # INPUTS:
    #   timeVec         : TimeVecType time vectory
    # OUTPUTS:
    #   fracDay         : See description.
    #
    # ---START CODE---
    fracDay = float(timeVec.day) + (float(timeVec.hour)/24.0) +
                                   (float(timeVec.minute)/1440.0) + 
                                   (float(timeVec.second)/86400.0)
    return fracDay
    # ---END CODE---
end

function julianDay(timeVec::TimeVecType)
    # FUNCTION NAME     : julianDay
    # DESCRIPTION       : Calculates the astronomical Julian day.
    #                     Algorithm taken from "Practical Astronomy
    #                     with your Calculator", Third Edition,
    #                     by Peter Duffett-Smith, Cambridge University
    #                     Press, 1992. This function is actually just
    #                     a feeder function for the conventional Julian
    #                     day calculation function.
    # INPUTS:
    #   timeVec         : TimeVecType time vector
    # OUTPUTS:
    #   jDay            : Julian day as a floating point number
    #
    # ---START CODE---
    return julianDay(timeVec.year,timeVec.month,calcFracDay(timeVec))
    # ---END CODE---
end

function julianDay(intYear::Int,intMonth::Int,fracDay::FloatingPoint)
    # FUNCTION NAME     : julianDay
    # DESCRIPTION       : Calculates the astronomical Julian day.
    #                     Algorithm taken from "Practical Astronomy
    #                     with your Calculator", Third Edition,
    #                     by Peter Duffett-Smith, Cambridge University
    #                     Press, 1992.
    # INPUTS:
    #   intYear         : Integer year value (YYYY)
    #   intMonth        : Integer month (1-12)
    #   fracDay         : Fractional days elapsed in month (0-28/29/30/31)
    # OUTPUTS:
    #   jDay            : Julian day as a floating point number
    #
    # ---START CODE---
    if (intMonth < 3)
        calcYear = intYear - 1
        calcMonth = intMonth + 12
    else
        calcYear = intYear
        calcMonth = intMonth
    end
    # Calculate A
    ATemp = mInt(float(intYear) / 100.0)

    # Gregorian calendar begins on 10 October 1582
    # All dates prior to this will be in the Julian calendar
    if intYear > 1582
        isGreg = true
    else
        if ((intYear==1582)&&(calcMonth>=10)&&(fracDay>= 15))
            isGreg = true
        else
            isGreg = false
        end
    end

    # Calculate B
    if isGreg
        BTemp = 2.0 - float(ATemp) + float(mInt(float(ATemp)/4.0))
    else
        BTemp = 0.0
    end

    # Calculate C
    if calcYear < 0
        X1Temp = (365.25 * calcYear) - 0.75
    else
        X1Temp = 365.25 * calcYear
    end
    CTemp = mInt(X1Temp)
    
    # Calculate D
    X1Temp = 30.6001 * float(calcMonth + 1)
    DTemp = mInt(X1Temp)

    # Put everything together to get the Julian Date
    jDay = float(BTemp+CTemp+DTemp) + fracDay + 1720994.5
    
    return jDay
    # ---END CODE---
end

function mInt(floatVal::FloatingPoint)
    # FUNCTION NAME     : mInt
    # DESCRIPTION       : Taken directly from GEOS-Chem.
    #                     Not sure why this is necessary.
    # INPUTS:
    #   floatVal        : Floating point input value
    # OUTPUTS:
    #   mIntVal         : Modified integer value
    #
    # ---START CODE---
    if (floatVal < 0.0)
        mIntVal = -itrunc(abs(floatVal))
    else
        mIntVal = itrunc(abs(floatVal))
    end
    return mIntVal
    # ---END CODE---
end

function calcTau(timeVec::TimeVecType)
    # FUNCTION NAME     : calcTau
    # DESCRIPTION       : Calculates the number of hours since
    #                     1985-01-01 00:00 as a floating point number.
    #                     This is actually just a feeder function to
    #                     to handle the TimeVecType datatype.
    # INPUTS:
    #   timeVec         : TimeVecType time vector
    # OUTPUTS:
    #   tauVal          : Hours since 1985-01-01 00:00
    # ---START CODE---
    return calcTau(timeVec.year,timeVec.month,timeVec.day,timeVec.hour,timeVec.minute,timeVec.second)
    # ---END CODE---
end

function calcTau(intYear::Int,intMonth::Int,intDay::Int,intHour::Int,intMin::Int,intSec::Int)
    # FUNCTION NAME     : calcTau
    # DESCRIPTION       : Calculates the number of hours since
    #                     1985-01-01 00:00 as a floating point number
    # INPUTS:
    #   intDate         : Integer date value (YYYYMMDD)
    #   intTime         : Integer time value (HHMMSS)
    # OUTPUTS:
    #   tauVal          : Hours since 1985-01-01 00:00
    #
    # ---START CODE---
    # Calculate number of days (as float) since start of month
    floatDay = float(intDay) + (float(intHour)/24.0) +
                               (float(intMin)/1440.0) + 
                               (float(intSec)/86400.0)
    # Calculate number of days (as float) since start of month
    floatDay = float(intDay) + (float(intHour)/24.0) +
                               (float(intMin)/1440.0) + 
                               (float(intSec)/86400.0)
    # Get Julian Date
    jDay = julianDay(intYear,intMonth,floatDay)
    jDay85 = 2556066.5 # Julian Date at 00:00 GMT 1985-01-01

    return ((jDay-jDay85)*24.0)
    # ---END CODE---
end

function calcTau(intDate::Int,intTime::Int)
    # FUNCTION NAME     : calcTau
    # DESCRIPTION       : Calculates the number of hours since
    #                     1985-01-01 00:00 as a floating point number
    # INPUTS:
    #   intDate         : Integer date value (YYYYMMDD)
    #   intTime         : Integer time value (HHMMSS)
    # OUTPUTS:
    #   tauVal          : Hours since 1985-01-01 00:00
    #
    # ---START CODE---
    # Split YYYYMMDD into year, month, day
    (intYear,intMonth,intDay) = parseIntDT(intDate)
    # As above for HHMMSS
    (intHour,intMin,intSec) = parseIntDT(intTime)
    return(calcTau(intYear,intMonth,intDay,intHour,intMin,intSec))
    # ---END CODE---
end

function parseIntDT(DTNum::Int)
    # FUNCTION NAME     : parseIntDT
    # DESCRIPTION       : Converts a YYYYMMDD integer
    #                     into integer year, month and day.
    #                     The same function can be used to
    #                     convert HHMMSS into hours, minutes
    #                     and seconds
    # INPUTS:
    #   DTNum           : YYYYMMDD or HHMMSS integer
    # OUTPUTS:
    #   intYH           : Year/Hour
    #   intMM           : Month/Minutes
    #   intDS           : Day/Seconds
    #
    # ---START CODE---

    intDS = mod(DTNum,100)
    intMM = mod(DTNum,10000) - intDS
    intYH = DTNum - (intMM + intDS)
    intMM = itrunc(intMM/100)
    intYH = itrunc(intYH/10000)
    
    return(intYH,intMM,intDS)
    # ---END CODE---
end

function expandDate(inString::String, dateNum::Int=0, timeNum::Int=0)
    # FUNCTION NAME     : expandDate
    # DESCRIPTION       : Expands a date formatting string,
    #                     replacing YYYY/YY with the year,
    #                     MM with the month, DD with the day,
    #                     hh with the hour, mm with the minute
    #                     and ss with the second
    # INPUTS:
    #   inString        : Input date formatting string
    # OUTPUTS:
    #   outString       : Formatted date string
    # Output variable declarations
    # Local variable declarations
    # ---START CODE---
    # Parse YYYYMMDD into integers
    if dateNum > 0
        (roundYear,roundMonth,roundDay) = parseIntDT(dateNum)
    else
       roundDay = 0
       roundMonth = 0
       roundYear = 0
    end
    strYear04 = @sprintf("%04i",roundYear)
    strYear02 = @sprintf("%02i",mod(roundYear,100));
    strMonth02= @sprintf("%02i",roundMonth);
    strDay02  = @sprintf("%02i",roundDay);
    # Parse HHMMSS into integers
    if timeNum > 0
        (roundHour,roundMin,roundSec) = parseIntDT(timeNum)
    else
        roundSec = 0
        roundMin = 0
        roundHour = 0
    end
    strHour02 = @sprintf("%02i",roundHour);
    strMin02  = @sprintf("%02i",roundMin);
    strSec02  = @sprintf("%02i",roundSec);
    # Sort out the rest
    outString = inString
    # Year (4 and 2 element version)
    outString = replace(outString,"YYYY",strYear04);
    outString = replace(outString,"YY",strYear02);
    # Month
    outString = replace(outString,"MM",strMonth02);
    # Day
    outString = replace(outString,"DD",strDay02);
    # Hour
    outString = replace(outString,"hh",strHour02);
    # Minute
    outString = replace(outString,"mm",strMin02);
    # Second
    outString = replace(outString,"ss",strSec02);
    return(outString)

    # ---END CODE---
end

function DTIntJZ(iYH::Int,iMM::Int=1,iDS::Int=1)
    # FUNCTION NAME     : DTIntJZ
    # DESCRIPTION       : Converts integer year, month and
    #                     day values into a YYYYMMDD integer.
    #                     Also works to create HHMMSS
    # INPUTS:
    #   iYH             : Integer year or hour
    #   iMM             : Integer month or minute
    #   iDS             : Integer day or second
    # OUTPUTS:
    #   outDT           : Date/time integer (YYYYMMDD/HHMMSS)
    #
    # Output variable declarations
    # Local variable declarations
    # ---START CODE---
    outDT = (iYH*10000) + (iMM*100) + iDS
    return(outDT)
    # ---END CODE---
end

function dateStr(iYear::Int,iMonth::Int,iDay::Int)
    # FUNCTION NAME     : dateStr
    # DESCRIPTION       : Converts integer year, month and
    #                     day values into a string.
    # INPUTS:
    #   iYear           : Integer year
    #   iMonth          : Integer month
    #   iDay            : Integer day
    # OUTPUTS:
    #   outDate         : Date string (YYYYMMDD)
    #
    # Output variable declarations
    # Local variable declarations
    # ---START CODE---
    outDate = @sprintf("%04d%02d%02d",iYear,iMonth,iDay)
    return(outDate)
    # ---END CODE---
end

function timeStr(iHour::Int,iMin::Int,iSec::Int)
    # FUNCTION NAME     : timeStr
    # DESCRIPTION       : Converts integer hour, minute and
    #                     second values into a string.
    # INPUTS:
    #   iHour           : Integer hour
    #   iMin            : Integer minute
    #   iSec            : Integer second
    # OUTPUTS:
    #   outTime         : Time string (HHMMSS)
    #
    # Output variable declarations
    # Local variable declarations
    # ---START CODE---
    outTime = @sprintf("%02d%02d%02d",iHour,iMin,iSec)
    return(outTime)
    # ---END CODE---
end

function calcA3Time(targTime::TimeVecType,firstRead::Bool=false)
    # FUNCTION NAME     : calcA3Time
    # DESCRIPTION       : Returns the time vector relevant
    #                     to the current A3 met data file.
    # INPUTS:
    #   targTime        : Target time
    #   firstRead       : Optional. Boolean - is this the first reading?
    # OUTPUTS:
    #   A3Time          : Time vector for A3 file
    #
    # ---START CODE---
    if firstRead
        # Compute the minutes elapsed in the 3-hour interval
        tOffset = mod(targTime.hour,3)*60 + targTime.minute
        # Now calculate the offset to the interval midpoint
        tOffset = 180 - (tOffset + 90)
    else
        # If not the first call, function should only be called
        # at the start of the 3-hour interval (ie 90 minutes
        # from the interval midpoint)
        tOffset = 90
    end
    outTime = copyTime(targTime)
    outTime.second = 0
    return calcTOffset(outTime,tOffset)
    # ---END CODE---
end

function calcA6Time(targTime::TimeVecType,firstRead::Bool=false)
    # FUNCTION NAME     : calcA6Time
    # DESCRIPTION       : Returns the time vector relevant
    #                     to the current A6 met data file.
    # INPUTS:
    #   targTime        : Target time
    #   firstRead       : Optional. Boolean - is this the first reading?
    # OUTPUTS:
    #   A6Time          : Time vector for A6 file
    #
    # ---START CODE---
    if firstRead
        # Compute the minutes elapsed in the 6-hour interval
        minOffset = mod(targTime.hour,6)*60 + targTime.minute
        # Now calculate the offset to the interval midpoint
        if minOffset < 180
            tOffset = -minOffset
        else
            tOffset = 360 - minOffset
        end
    else
        # If not the first call, function should only be called
        # at the start of the 6-hour interval (ie 180 minutes
        # from the interval midpoint)
        tOffset = 180
    end
    outTime = copyTime(targTime)
    outTime.second = 0
    return calcTOffset(outTime,tOffset)
    # ---END CODE---
end

function calcI6Time(targTime::TimeVecType,firstRead::Bool=false)
    # FUNCTION NAME     : calcI6Time
    # DESCRIPTION       : Returns the time vector relevant
    #                     to the current I6 met data file.
    # INPUTS:
    #   targTime        : Target time
    #   firstRead       : Optional. Boolean - is this the first reading?
    # OUTPUTS:
    #   I6Time          : Time vector for I6 file
    #
    # ---START CODE---
    if firstRead
        # Compute the minutes elapsed in the 6-hour interval
        tOffset = mod(targTime.hour,6)*60 + targTime.minute
        # Now calculate the offset to the nearest midpoint
        tOffset = -tOffset
    else
        # With I6, some confusion - GC has both a "FIRST_I6_TIME" function
        # and a function "I6_TIME" which acts differently depending on
        # whether it is being called for the first time. However, the only
        # difference occurs if the first function call occurs away from
        # a sample point (ie MOD(H,6) != 0)
        if mod(targTime.hour,6) != 0
            tOffset = 360 - ((mod(targTime.hour,6)*60) + targTime.minute)
        else
            tOffset = 360
        end
    end
    outTime = copyTime(targTime)
    outTime.second = 0
    return calcTOffset(outTime,tOffset)
    # ---END CODE---
end

function calcTDiff(plusTime::TimeVecType,minusTime::TimeVecType)
    # FUNCTION NAME     : calcTDiff
    # DESCRIPTION       : Calculates the difference in minutes
    #                     appropriate to (plusTime - minusTime)
    # INPUTS:
    #   plusTime        : Minuend TimeVecType
    #   minusTime       : Subtrahend TimeVecType
    # OUTPUTS:
    #   tDiff           : Difference in minutes
    #
    # ---START CODE---
    tDiff = 60.0*(calcTau(plusTime)-calcTau(minusTime))
    return tDiff
    # ---END CODE---
end

function calcTOffset(baseTime::TimeVecType,plusMins::Int)
    # FUNCTION NAME     : calcTOffset
    # DESCRIPTION       : Calculates the time plus or minus
    #                     a whole number of minutes
    # INPUTS:
    #   baseTime        : Time vector containing the base time
    #   plusMins        : Number of minutes to add to baseTime
    # OUTPUTS:
    #   offsetTime      : Time vector with offset applied
    #
    # ---START CODE---
    # Apply offset directly to Julian day, then reconvert to a time vector
    currJDay = julianDay(baseTime.year,baseTime.month,calcFracDay(baseTime))
    offsetJDay = currJDay + (float(plusMins)/1440.0)
    return TimeVecType(offsetJDay)
    # ---END CODE---
end

function addTStep(timeData::JZTimeType)
    # FUNCTION NAME     : addTStep
    # DESCRIPTION       : Increments the current time by
    #                     the given time step
    # INPUTS:
    #   timeData        : Simulation timing data
    #   tStep           : Time step (minutes)
    # OUTPUTS:
    #   [Modified timeData]
    #
    # ---START CODE---
    timeData.tElapsed += timeData.tStepDyn
    timeData.tCurr = calcTOffset(timeData.tCurr,timeData.tStepDyn)
    return(timeData)
    # ---END CODE---
end

# ---END TIME FUNCTIONS
# ---START MISC FUNCTIONS
function calcTracMass(tracerData::JZTracGridType,metState::JZMetType)
    # FUNCTION NAME     : calcTracMass
    # DESCRIPTION       : Calculates the sum total of tracer
    #                     in the grid (for every tracer) in kg
    # INPUTS:
    #   tracerData      : Current tracer grids
    #   metState        : Current meteorology
    #   JZFiles         : A generic array of file IOStreams
    # OUTPUTS:
    #   tracerSum       : Vector of mass totals
    #
    # ---START CODE---
    nTrc = tracerData.nTracers
    dMR = tracerData.mixRatio
    dAM = metState.airMass
    relWeight = tracerData.relWeight

    tracerSum = zeros(nTrc)
    @sync begin
        [tracerSum += @fetchfrom iWorker begin
            cMR = localpart(dMR)
            nLonLocal = size(cMR,1)
            nLatLocal = size(cMR,2)
            nLev = size(cMR,3)
            nTrc = size(cMR,4)
            cAM = localpart(dAM)
            miniSum = zeros(nTrc)
            relWeightLocal = remotecall_fetch(myid(),()->relWeight)
            for iTrc in 1:nTrc
                for iLon in 1:nLonLocal
                    for iLat in 1:nLatLocal
                        for iLev in 1:nLev
                            miniSum[iTrc] += cAM[iLon,iLat,iLev]*cMR[iLon,iLat,iLev,iTrc]/relWeightLocal[iTrc]
                        end
                    end
                end
            end
            miniSum
        end
        for iWorker in workers()]
    end
    # ---END CODE---
    return tracerSum
end

function cleanupJZ(metState::JZMetType,JZFiles::Array)
    # FUNCTION NAME     : cleanupJZ
    # DESCRIPTION       : Runs cleanup operations to ensure
    #                     that J-Zephyr exits without leaving
    #                     possible memory leaks etc.
    # INPUTS:
    #   metState        : Current meteorology
    #   JZFiles         : A generic array of file IOStreams
    # OUTPUTS:
    #   exitFlag        : 0 if no problems, 1 otherwise
    #
    # ---START CODE---

    # Many of the following file streams may already be closed,
    # but the commands should still execute without error

    # Close the generic JZFiles array
    for iFile = 1:length(JZFiles)
        close(JZFiles[iFile])
    end

    # Close met file IOStreams
    close(metState.I6File)
    close(metState.A6File)
    close(metState.A3File)

    # Close output files
    #close(diagData.tracerFile)
    #close(diagData.tEdgeFile)

    # Nothing in here that could throw an error
    exitFlag = 0
    return(exitFlag)
    # ---END CODE---
end

# ---END MISC FUNCTIONS---

# ---END JULIA ZEPHYR---

function nullFunc()
    return(nothing)
end

#function sampleFunc(inVar::Type)
#    # FUNCTION NAME     : sampleFunc
#    # DESCRIPTION       : Function description, line 1...
#    #                     and line 2. 14 lines.
#    # INPUTS:
#    #   inVar           : Input description
#    # OUTPUTS:
#    #   varName         : Output description
#    #
#    # ---START CODE---
#    varName = localVar + inVar
#    return(varName)
#    # ---END CODE---
#end
